"""Proximal Policy Optimization (PPO) algorithm for reinforcement learning."""
import copy
import os
import time

from collections.abc import Sequence as SequenceCollection
from multiprocessing.dummy import Pool

import numpy as np
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.optimizers import Adam, Optimizer
from typing import Any, Dict, List, Optional, Union
try:
    from deepchem.rl import Environment, Policy
    import torch
    import torch.nn as nn
    has_pytorch = True
except:
    has_pytorch = False


class PPOLoss(object):
    """This class computes the loss function for PPO.

    Example
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> import torch
    >>> import torch.nn.functional as F
    >>> outputs = [torch.tensor([[0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02,0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2]]), torch.tensor([0.], requires_grad = True)]
    >>> labels = np.array([[0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]], dtype = np.float32)
    >>> discount = np.array([-1.0203744, -0.02058018, 0.98931295, 2.009407, 1.019603, 0.01980097, -0.9901, 0.01, -1. , 0. ], dtype=np.float32)
    >>> advantage = np.array([-1.0203744 ,-0.02058018, 0.98931295, 2.009407, 1.019603, 0.01980097, -0.9901 ,0.01 ,-1. , 0.], dtype = np.float32)
    >>> old_prob = np.array([0.28183755, 0.95147914, 0.87922776, 0.8037652 , 0.11757819, 0.271103  , 0.21057394, 0.78721744, 0.6545527 , 0.8832647 ], dtype=np.float32)
    >>> loss = PPOLoss(value_weight = 1.0, entropy_weight = 0.01, clipping_width = 0.2, action_prob_index = 0, value_index = 1)
    >>> loss_val = loss(outputs, [labels], [discount, advantage, old_prob])
    >>> loss_val
    tensor(1.0761, grad_fn=<SubBackward0>)
    """

    def __init__(self, value_weight: float, entropy_weight: float,
                 clipping_width: float, action_prob_index: int,
                 value_index: int):
        self.value_weight: float = value_weight
        self.entropy_weight: float = entropy_weight
        self.clipping_width: float = clipping_width
        self.action_prob_index: int = action_prob_index
        self.value_index: int = value_index

    def __call__(self, outputs: List[torch.Tensor], labels: List[np.ndarray],
                 weights: List[np.ndarray]):
        prob: torch.Tensor = outputs[self.action_prob_index]
        value: torch.Tensor = outputs[self.value_index]
        reward, advantage_array, old_prob = weights
        action: torch.Tensor = torch.from_numpy(labels[0])
        advantage: torch.Tensor = torch.unsqueeze(
            torch.from_numpy(advantage_array), 1)
        machine_eps = torch.finfo(torch.float32).eps
        prob = prob + machine_eps
        old_prob = old_prob + machine_eps
        ratio: torch.Tensor = torch.sum(action * prob,
                                        1) / torch.from_numpy(old_prob)
        clipped_ratio: torch.Tensor = torch.clamp(ratio,
                                                  1 - self.clipping_width,
                                                  1 + self.clipping_width)
        policy_loss: torch.Tensor = -torch.mean(
            torch.minimum(ratio * advantage, clipped_ratio * advantage))
        value_loss: torch.Tensor = torch.mean(
            torch.square(torch.from_numpy(reward) - value))
        entropy: torch.Tensor = -torch.mean(torch.sum(prob * torch.log(prob),
                                                      1))
        return policy_loss + self.value_weight * value_loss - self.entropy_weight * entropy


class PPO(object):
    """
    Implements the Proximal Policy Optimization (PPO) algorithm for reinforcement learning.

    The algorithm is described in Schulman et al, "Proximal Policy Optimization Algorithms"
    (https://openai-public.s3-us-west-2.amazonaws.com/blog/2017-07/ppo/ppo-arxiv.pdf).
    This class requires the policy to output two quantities: a vector giving the probability of
    taking each action, and an estimate of the value function for the current state.  It
    optimizes both outputs at once using a loss that is the sum of three terms:

    1. The policy loss, which seeks to maximize the discounted reward for each action.
    2. The value loss, which tries to make the value estimate match the actual discounted reward
        that was attained at each step.
    3. An entropy term to encourage exploration.

    This class only supports environments with discrete action spaces, not continuous ones.  The
    "action" argument passed to the environment is an integer, giving the index of the action to perform.

    This class supports Generalized Advantage Estimation as described in Schulman et al., "High-Dimensional
    Continuous Control Using Generalized Advantage Estimation" (https://arxiv.org/abs/1506.02438).
    This is a method of trading off bias and variance in the advantage estimate, which can sometimes
    improve the rate of convergance.  Use the advantage_lambda parameter to adjust the tradeoff.

    This class supports Hindsight Experience Replay as described in Andrychowicz et al., "Hindsight
    Experience Replay" (https://arxiv.org/abs/1707.01495).  This is a method that can enormously
    accelerate learning when rewards are very rare.  It requires that the environment state contains
    information about the goal the agent is trying to achieve.  Each time it generates a rollout, it
    processes that rollout twice: once using the actual goal the agent was pursuing while generating
    it, and again using the final state of that rollout as the goal.  This guarantees that half of
    all rollouts processed will be ones that achieved their goals, and hence received a reward.

    To use this feature, specify use_hindsight=True to the constructor.  The environment must have
    a method defined as follows:

    def apply_hindsight(self, states, actions, goal):
        ...
        return new_states, rewards

    The method receives the list of states generated during the rollout, the action taken for each one,
    and a new goal state.  It should generate a new list of states that are identical to the input ones,
    except specifying the new goal.  It should return that list of states, and the rewards that would
    have been received for taking the specified actions from those states.  The output arrays may be
    shorter than the input ones, if the modified rollout would have terminated sooner.

    Example
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> import torch
    >>> import torch.nn.functional as F
    >>> from deepchem.rl.torch_rl import PPO
    >>> class RouletteEnvironment(dc.rl.Environment):
    ...     def __init__(self):
    ...         super(RouletteEnvironment, self).__init__([(1,)], 38)
    ...         self._state = [np.array([0])]
    ...     def step(self, action):
    ...         if action == 37:
    ...             self._terminated = True  # Walk away.
    ...             return 0.0
    ...         wheel = np.random.randint(37)
    ...         if wheel == 0:
    ...             if action == 0:
    ...                 return 35.0
    ...             return -1.0
    ...         if action != 0 and wheel % 2 == action % 2:
    ...             return 1.0
    ...         return -1.0
    ...     def reset(self):
    ...         self._terminated = False
    >>> class TestPolicy(dc.rl.Policy):
    ...     def __init__(self):
    ...         super(TestPolicy, self).__init__(['action_prob', 'value'])
    ...     def create_model(self, **kwargs):
    ...         class TestModel(nn.Module):
    ...             def __init__(self):
    ...                 super(TestModel, self).__init__()
    ...                 self.action = nn.Parameter(torch.ones(env.n_actions, dtype=torch.float32))
    ...                 self.value = nn.Parameter(torch.tensor([0.0], dtype=torch.float32))
    ...             def forward(self, inputs):
    ...                 prob = F.softmax(torch.reshape(self.action, (-1, env.n_actions)))
    ...                 return prob, self.value
    ...         return TestModel()
    >>> env = RouletteEnvironment()
    >>> ppo = PPO(env, TestPolicy(), max_rollout_length=20, optimization_epochs=8, optimizer=Adam(learning_rate=0.003))
    >>> ppo.fit(100000)
    >>> action_prob, value = ppo.predict([[0]])
    """

    def __init__(self,
                 env: Environment,
                 policy: Policy,
                 max_rollout_length: int = 20,
                 optimization_rollouts: int = 8,
                 optimization_epochs: int = 4,
                 batch_size: int = 64,
                 clipping_width: float = 0.2,
                 discount_factor: float = 0.99,
                 advantage_lambda: float = 0.98,
                 value_weight: float = 1.0,
                 entropy_weight: float = 0.01,
                 optimizer: Optional[Optimizer] = None,
                 model_dir: Optional[str] = None,
                 use_hindsight: bool = False,
                 device: Optional[torch.device] = None) -> None:
        """Create an object for optimizing a policy.

        Parameters
        ----------
        env: Environment
            the Environment to interact with
        policy: Policy
            the Policy to optimize.  It must have outputs with the names 'action_prob'
            and 'value', corresponding to the action probabilities and value estimate
        max_rollout_length: int
            the maximum length of rollouts to generate
        optimization_rollouts: int
            the number of rollouts to generate for each iteration of optimization
        optimization_epochs: int
            the number of epochs of optimization to perform within each iteration
        batch_size: int
            the batch size to use during optimization.  If this is 0, each rollout will be used as a
            separate batch.
        clipping_width: float
            in computing the PPO loss function, the probability ratio is clipped to the range
            (1-clipping_width, 1+clipping_width)
        discount_factor: float
            the discount factor to use when computing rewards
        advantage_lambda: float
            the parameter for trading bias vs. variance in Generalized Advantage Estimation
        value_weight: float
            a scale factor for the value loss term in the loss function
        entropy_weight: float
            a scale factor for the entropy term in the loss function
        optimizer: Optimizer
            the optimizer to use.  If None, a default optimizer is used.
        model_dir: str
            the directory in which the model will be saved.  If None, a temporary directory will be created.
        use_hindsight: bool
            if True, use Hindsight Experience Replay
        """
        self._env: Environment = env
        self._policy: Policy = policy
        self.max_rollout_length: int = max_rollout_length
        self.optimization_rollouts: int = optimization_rollouts
        self.optimization_epochs: int = optimization_epochs
        self.batch_size: int = batch_size
        self.clipping_width: float = clipping_width
        self.discount_factor: float = discount_factor
        self.advantage_lambda: float = advantage_lambda
        self.value_weight: float = value_weight
        self.entropy_weight: float = entropy_weight
        self.use_hindsight: bool = use_hindsight
        self._state_is_list: bool = isinstance(env.state_shape[0],
                                               SequenceCollection)
        if optimizer is None:
            self._optimizer: Optimizer = Adam(learning_rate=0.001,
                                              beta1=0.9,
                                              beta2=0.999)
        else:
            self._optimizer = optimizer
        output_names: List[str] = policy.output_names
        self._value_index: int = output_names.index('value')
        self._action_prob_index: int = output_names.index('action_prob')
        self._rnn_final_state_indices: List[int] = [
            i for i, n in enumerate(output_names) if n == 'rnn_state'
        ]
        self._rnn_states: List[np.ndarray] = policy.rnn_initial_states
        if len(self._rnn_states) > 0 and batch_size != 0:
            raise ValueError(
                'Cannot batch rollouts when the policy contains a recurrent layer.  Set batch_size to 0.'
            )
        self._model: TorchModel = self._build_model(model_dir)
        if device is None:
            if torch.cuda.is_available():
                device = torch.device('cuda')
            elif torch.backends.mps.is_available():
                device = torch.device('mps')
            else:
                device = torch.device('cpu')
        self.device: torch.device = device
        self._model.model = self._model.model.to(device)

    def _build_model(self, model_dir: Optional[str]) -> TorchModel:
        """Construct a TorchModel containing the policy and loss calculations."""
        policy_model: nn.Module = self._policy.create_model()
        loss: PPOLoss = PPOLoss(self.value_weight, self.entropy_weight,
                                self.clipping_width, self._action_prob_index,
                                self._value_index)
        model: TorchModel = TorchModel(policy_model,
                                       loss,
                                       batch_size=self.max_rollout_length,
                                       model_dir=model_dir,
                                       optimize=self._optimizer)
        model._ensure_built()
        return model

    def fit(self,
            total_steps: int,
            max_checkpoints_to_keep: int = 5,
            checkpoint_interval: int = 600,
            restore: bool = False) -> None:
        """Train the policy.

        Parameters
        ----------
        total_steps: int
            the total number of time steps to perform on the environment, across all rollouts
            on all threads
        max_checkpoints_to_keep: int
            the maximum number of checkpoint files to keep.  When this number is reached, older
            files are deleted.
        checkpoint_interval: float
            the time interval at which to save checkpoints, measured in seconds
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        """
        step_count: int = 0
        workers: List[_Worker] = []
        for i in range(self.optimization_rollouts):
            workers.append(_Worker(self, i))
        if restore:
            self.restore()
        pool = Pool()
        self._model.model.train()
        checkpoint_time: float = time.time()
        while step_count < total_steps:
            # Have the worker threads generate the rollouts for this iteration.

            rollouts = []
            pool.map(lambda x: rollouts.extend(x.run()), workers)

            # Perform optimization.

            for epoch in range(self.optimization_epochs):
                if self.batch_size == 0:
                    batches = rollouts
                else:
                    batches = self._iter_batches(rollouts)
                for batch in batches:
                    initial_rnn_states, state_arrays, discounted_rewards, actions_matrix, action_prob, advantages = batch

                    # Build the inputs and run the optimizer.

                    state_arrays = [np.stack(s) for s in state_arrays]
                    inputs = state_arrays + [
                        np.expand_dims(s, axis=0) for s in initial_rnn_states
                    ]
                    self._apply_gradients(inputs, actions_matrix,
                                          discounted_rewards, advantages,
                                          action_prob)

            # Update the number of steps taken so far and perform checkpointing.

            new_steps = sum(len(r[3]) for r in rollouts)
            if self.use_hindsight:
                new_steps = int(new_steps / 2)
            step_count += new_steps
            if step_count >= total_steps or time.time(
            ) >= checkpoint_time + checkpoint_interval:
                self.save_checkpoint(max_checkpoints_to_keep)
                checkpoint_time = time.time()

    def _apply_gradients(self, inputs: List[np.ndarray],
                         actions_matrix: np.ndarray,
                         discounted_rewards: np.ndarray, advantages: np.ndarray,
                         action_prob: np.ndarray):
        """Compute the gradient of the loss function for a batch and update the model."""
        self._model._pytorch_optimizer.zero_grad()
        outputs: List[torch.Tensor] = self._model.model(inputs)
        loss: torch.Tensor = self._model._loss_fn(
            outputs, [actions_matrix],
            [discounted_rewards, advantages, action_prob])
        loss.backward()
        self._model._pytorch_optimizer.step()
        if self._model._lr_schedule is not None:
            self._model._lr_schedule.step()

    def _iter_batches(self, rollouts: List[np.ndarray]) -> Any:
        """Given a set of rollouts, merge them into batches for optimization."""

        # Merge all the rollouts into a single set of arrays.

        state_arrays: List[np.ndarray] = []
        for i in range(len(rollouts[0][1])):
            state_arrays.append(np.concatenate([r[1][i] for r in rollouts]))
        discounted_rewards: np.ndarray = np.concatenate(
            [r[2] for r in rollouts])
        actions_matrix: np.ndarray = np.concatenate([r[3] for r in rollouts])
        action_prob: np.ndarray = np.concatenate([r[4] for r in rollouts])
        advantages: np.ndarray = np.concatenate([r[5] for r in rollouts])
        total_length: int = len(discounted_rewards)

        # Iterate slices.

        start = 0
        while start < total_length:
            end: int = min(start + self.batch_size, total_length)
            batch: List[Any] = [[]]
            batch.append([s[start:end] for s in state_arrays])
            batch.append(discounted_rewards[start:end])
            batch.append(actions_matrix[start:end])
            batch.append(action_prob[start:end])
            batch.append(advantages[start:end])
            start = end
            yield batch

    def predict(self,
                state: np.ndarray,
                use_saved_states: bool = True,
                save_states: bool = True) -> List[np.ndarray]:
        """Compute the policy's output predictions for a state.

        If the policy involves recurrent layers, this method can preserve their internal
        states between calls.  Use the use_saved_states and save_states arguments to specify
        how it should behave.

        Parameters
        ----------
        state: array or list of arrays
            the state of the environment for which to generate predictions
        use_saved_states: bool
            if True, the states most recently saved by a previous call to predict() or select_action()
            will be used as the initial states.  If False, the internal states of all recurrent layers
            will be set to the initial values defined by the policy before computing the predictions.
        save_states: bool
            if True, the internal states of all recurrent layers at the end of the calculation
            will be saved, and any previously saved states will be discarded.  If False, the
            states at the end of the calculation will be discarded, and any previously saved
            states will be kept.

        Returns
        -------
        the array of action probabilities, and the estimated value function
        """
        results: List[np.ndarray] = self._predict_outputs(
            state, use_saved_states, save_states)
        return [
            results[i] for i in (self._action_prob_index, self._value_index)
        ]

    def select_action(self,
                      state: List[np.ndarray],
                      deterministic: bool = False,
                      use_saved_states: bool = True,
                      save_states: bool = True) -> int:
        """Select an action to perform based on the environment's state.

        If the policy involves recurrent layers, this method can preserve their internal
        states between calls.  Use the use_saved_states and save_states arguments to specify
        how it should behave.

        Parameters
        ----------
        state: array or list of arrays
            the state of the environment for which to select an action
        deterministic: bool
            if True, always return the best action (that is, the one with highest probability).
            If False, randomly select an action based on the computed probabilities.
        use_saved_states: bool
            if True, the states most recently saved by a previous call to predict() or select_action()
            will be used as the initial states.  If False, the internal states of all recurrent layers
            will be set to the initial values defined by the policy before computing the predictions.
        save_states: bool
            if True, the internal states of all recurrent layers at the end of the calculation
            will be saved, and any previously saved states will be discarded.  If False, the
            states at the end of the calculation will be discarded, and any previously saved
            states will be kept.

        Returns
        -------
        the index of the selected action
        """
        outputs: List[np.ndarray] = self._predict_outputs(
            state, use_saved_states, save_states)
        return self._select_action_from_outputs(outputs, deterministic)

    def restore(self, strict: Optional[bool] = True) -> None:
        """Reload the model parameters from the most recent checkpoint file."""
        last_checkpoint: Union[List[str], str] = sorted(
            self.get_checkpoints(self._model.model_dir))
        if last_checkpoint is None:
            raise ValueError('No checkpoint found')
        last_checkpoint = last_checkpoint[0]
        data: Any = torch.load(last_checkpoint, map_location=self.device)
        self._model.model.load_state_dict(data['model_state_dict'],
                                          strict=strict)
        self._model._pytorch_optimizer.load_state_dict(
            data['optimizer_state_dict'])
        self._model._global_step = data['global_step']

    def _predict_outputs(self, state, use_saved_states: bool,
                         save_states: bool) -> List[np.ndarray]:
        """Compute a set of outputs for a state. """
        if not self._state_is_list:
            state = [state]
        if use_saved_states:
            state = state + list(self._rnn_states)
        else:
            state = state + list(self._policy.rnn_initial_states)
        inputs: List[np.ndarray] = [np.expand_dims(s, axis=0) for s in state]
        results = self._model.model(inputs)
        results = [r.detach().numpy() for r in results]
        if save_states:
            self._rnn_states = [
                np.squeeze(results[i], 0) for i in self._rnn_final_state_indices
            ]
        return results

    def _select_action_from_outputs(self, outputs: List[np.ndarray],
                                    deterministic: bool) -> int:
        """Given the policy outputs, select an action to perform."""
        action_prob: np.ndarray = outputs[self._action_prob_index]
        if deterministic:
            return int(action_prob.argmax())
        else:
            action_prob = action_prob.flatten()
            return np.random.choice(np.arange(len(action_prob)), p=action_prob)

    def save_checkpoint(self,
                        max_checkpoints_to_keep: int = 5,
                        model_dir: Optional[str] = None) -> None:
        """Save a checkpoint to disk.

        Usually you do not need to call this method, since fit() saves checkpoints
        automatically.  If you have disabled automatic checkpointing during fitting,
        this can be called to manually write checkpoints.

        Parameters
        ----------
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        model_dir: str, default None
            Model directory to save checkpoint to. If None, revert to self.model_dir
        """
        if model_dir is None:
            model_dir = self._model.model_dir
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        # Save the checkpoint to a file.

        data: Dict[str, Any] = {
            'model_state_dict': self._model.model.state_dict(),
            'optimizer_state_dict': self._model._pytorch_optimizer.state_dict(),
            'global_step': self._model._global_step
        }
        temp_file: str = os.path.join(model_dir, 'temp_checkpoint.pt')
        torch.save(data, temp_file)

        # Rename and delete older files.

        paths: List[str] = [
            os.path.join(model_dir, 'checkpoint%d.pt' % (i + 1))
            for i in range(max_checkpoints_to_keep)
        ]
        if os.path.exists(paths[-1]):
            os.remove(paths[-1])
        for i in reversed(range(max_checkpoints_to_keep - 1)):
            if os.path.exists(paths[i]):
                os.rename(paths[i], paths[i + 1])
        os.rename(temp_file, paths[0])

    def get_checkpoints(self, model_dir: Optional[str] = None) -> List[str]:
        """Get a list of all available checkpoint files.

        Parameters
        ----------
        model_dir: str, default None
            Directory to get list of checkpoints from. Reverts to self.model_dir if None

        """
        if model_dir is None:
            model_dir = self._model.model_dir
        files: List[str] = sorted(os.listdir(model_dir))
        files = [
            f for f in files if f.startswith('checkpoint') and f.endswith('.pt')
        ]
        return [os.path.join(model_dir, f) for f in files]


class _Worker(object):
    """A Worker object is created for each training thread."""

    def __init__(self, ppo, index):
        self.ppo = ppo
        self.index = index
        self.scope = 'worker%d' % index
        self.env = copy.deepcopy(ppo._env)
        self.env.reset()
        self.model = ppo._build_model(None)
        self.rnn_states = ppo._policy.rnn_initial_states

    def run(self):
        rollouts = []
        local_params = list(self.model.model.parameters())
        global_params = list(self.ppo._model.model.parameters())

        # Copy parameters from global_model to local_model
        with torch.no_grad():
            for local_param, global_param in zip(local_params, global_params):
                local_param.data.copy_(global_param.detach())
        initial_rnn_states = self.rnn_states
        states, actions, action_prob, rewards, values = self.create_rollout()
        rollouts.append(
            self.process_rollout(states, actions, action_prob, rewards, values,
                                 initial_rnn_states))
        if self.ppo.use_hindsight:
            rollouts.append(
                self.process_rollout_with_hindsight(states, actions,
                                                    initial_rnn_states))
        return rollouts

    def create_rollout(self):
        """Generate a rollout."""
        states = []
        action_prob = []
        actions = []
        rewards = []
        values = []

        # Generate the rollout.

        for i in range(self.ppo.max_rollout_length):
            if self.env.terminated:
                break
            state = self.env.state
            states.append(state)
            results = self.model.model(
                self._create_model_inputs(state, self.rnn_states))
            results = [r.detach().numpy() for r in results]
            value = results[self.ppo._value_index]
            probabilities = np.squeeze(results[self.ppo._action_prob_index])
            self.rnn_states = [
                np.squeeze(results[i], 0)
                for i in self.ppo._rnn_final_state_indices
            ]
            action = self.ppo._select_action_from_outputs(results, False)
            actions.append(action)
            action_prob.append(probabilities[action])
            values.append(float(value))
            rewards.append(self.env.step(action))

        # Compute an estimate of the reward for the rest of the episode.

        if not self.env.terminated:
            results = self.model.model(
                self._create_model_inputs(self.env.state, self.rnn_states))
            final_value = self.ppo.discount_factor * results[
                self.ppo._value_index].detach().numpy()[0]
            try:
                final_value = final_value.tolist()[0]
            except:
                pass
        else:
            final_value = 0.0
        values.append(final_value)
        if self.env.terminated:
            self.env.reset()
            self.rnn_states = self.ppo._policy.rnn_initial_states
        return states, np.array(actions, dtype=np.int32), np.array(
            action_prob, dtype=np.float32), np.array(
                rewards, dtype=np.float32), np.array(values, dtype=np.float32)

    def process_rollout(self, states, actions, action_prob, rewards, values,
                        initial_rnn_states):
        """Construct the arrays needed for training."""

        # Compute the discounted rewards and advantages.

        discounted_rewards = rewards.copy()
        discounted_rewards[-1] += values[-1]
        advantages = rewards - values[:-1] + self.ppo.discount_factor * np.array(
            values[1:])
        for j in range(len(rewards) - 1, 0, -1):
            discounted_rewards[
                j - 1] += self.ppo.discount_factor * discounted_rewards[j]
            advantages[
                j -
                1] += self.ppo.discount_factor * self.ppo.advantage_lambda * advantages[
                    j]

        # Convert the actions to one-hot.

        n_actions = self.env.n_actions
        actions_matrix = []
        for action in actions:
            a = np.zeros(n_actions, np.float32)
            a[action] = 1.0
            actions_matrix.append(a)
        actions_matrix = np.array(actions_matrix, dtype=np.float32)

        # Rearrange the states into the proper set of arrays.

        if self.ppo._state_is_list:
            state_arrays = [[] for i in range(len(self.env.state_shape))]
            for state in states:
                for j in range(len(state)):
                    state_arrays[j].append(state[j])
        else:
            state_arrays = [states]

        # Return the processed arrays.

        return (initial_rnn_states, state_arrays, discounted_rewards,
                actions_matrix, action_prob, advantages)

    def process_rollout_with_hindsight(self, states, actions,
                                       initial_rnn_states):
        """Create a new rollout by applying hindsight to an existing one, then process it."""
        hindsight_states, rewards = self.env.apply_hindsight(
            states, actions, states[-1])
        if self.ppo._state_is_list:
            state_arrays = [[] for i in range(len(self.env.state_shape))]
            for state in hindsight_states:
                for j in range(len(state)):
                    state_arrays[j].append(state[j])
        else:
            state_arrays = [hindsight_states]
        state_arrays += initial_rnn_states
        state_arrays = [np.stack(s) for s in state_arrays]
        inputs = state_arrays + [
            np.expand_dims(s, axis=0) for s in initial_rnn_states
        ]
        outputs = self.model.model(inputs)
        values = outputs[self.ppo._value_index].detach().numpy()
        values = np.append(values.flatten(), 0.0)
        probabilities = outputs[self.ppo._action_prob_index].detach().numpy()
        actions = actions[:len(rewards)]
        action_prob = probabilities[np.arange(len(actions)), actions]
        return self.process_rollout(hindsight_states, actions, action_prob,
                                    np.array(rewards, dtype=np.float32),
                                    np.array(values, dtype=np.float32),
                                    initial_rnn_states)

    def _create_model_inputs(self, state, rnn_states):
        """Create the inputs to the model for use during a rollout."""
        if not self.ppo._state_is_list:
            state = [state]
        state = state + rnn_states
        return [np.expand_dims(s, axis=0) for s in state]



import os
import deepchem as dc
import numpy as np


def test_batch_coulomb_matrix_features():
    # Get Data
    current_dir = os.path.dirname(os.path.abspath(__file__))
    dataset_file = os.path.join(current_dir, 'assets/qm9_mini.sdf')
    TASKS = ["alpha", "homo"]
    loader = dc.data.SDFLoader(tasks=TASKS,
                               featurizer=dc.feat.CoulombMatrix(29),
                               sanitize=True)
    data = loader.create_dataset(dataset_file, shard_size=100)
    inputs = dc.utils.batch_utils.batch_coulomb_matrix_features(data.X)

    # Checks that all atoms exits in array
    assert inputs[0].shape == (149,)

    # Checks shape of gaussian distance
    assert inputs[1].shape == (1215, 100)

    # Checks all molecule membership exist
    for i in range(0, 21):
        if i not in inputs[2]:
            raise AssertionError("All molecules not in the matrix")

    # Check Distance Membership shape
    assert inputs[3].shape == (1215,)
    assert inputs[4].shape == (1215,)


def test_batch_elements():
    # Prepare Data
    inputs = [[i, i**2, i**3] for i in range(10)]
    # Run
    output = list(dc.utils.batch_utils.batch_elements(inputs, 3))
    assert output == [[[0, 0, 0], [1, 1, 1], [2, 4, 8]],
                      [[3, 9, 27], [4, 16, 64], [5, 25, 125]],
                      [[6, 36, 216], [7, 49, 343], [8, 64, 512]], [[9, 81,
                                                                    729]]]


def test_create_input_array():
    # Prepare Data
    inputs = [["a", "b"], ["b", "b", "b"]]
    input_dict = {"c": 0, "a": 1, "b": 2}

    # Inputs property
    max_length = max([len(x) for x in inputs])

    # Without reverse input
    output_1 = dc.utils.batch_utils.create_input_array(inputs, max_length,
                                                       False, 2, input_dict,
                                                       "c")

    assert output_1.shape == (2, max_length + 1)
    assert np.allclose(output_1, np.array([[1., 2., 0., 0.], [2., 2., 2., 0.]]))

    # With revercse input
    output_2 = dc.utils.batch_utils.create_input_array(inputs, max_length, True,
                                                       2, input_dict, "c")

    assert output_2.shape == (2, max_length + 1)
    assert np.allclose(output_2, np.array([[2., 1., 0., 0.], [2., 2., 2., 0.]]))


def test_create_output_array():
    # Prepare Data
    inputs = [["a", "b"], ["b", "b", "b"]]
    output_dict = {"c": 0, "a": 1, "b": 2}

    # Inputs property
    max_length = max([len(x) for x in inputs])

    output = dc.utils.batch_utils.create_output_array(inputs, max_length, 2,
                                                      output_dict, "c")

    assert output.shape == (2, 3)
    assert np.allclose(output, np.array([[1., 2., 0.], [2., 2., 2.]]))



"""
Contains implementations of layers used in ChemCeption.

ChemCeption is a deep learning model designed for molecular property prediction using convolutional neural networks (CNNs).
It adapts the Inception-ResNet architecture to process molecular images effectively.

References:
- ChemCeption: https://arxiv.org/abs/1710.02238
"""
import torch
import torch.nn as nn


class Stem(nn.Module):
    """
    Implements the Stem Layer as defined in https://arxiv.org/abs/1710.02238.

    This layer serves as the initial processing block in ChemCeption,
    downsampling input images to reduce computational complexity
    before they pass through deeper network layers. The convolutional layer
    with stride 2 helps in feature extraction while reducing spatial dimensions.

    Examples
    --------
    >>> import numpy as np
    >>> import torch
    >>> from deepchem.models.torch_models.chemnet_layers import Stem
    >>> in_channels = 3
    >>> out_channels = 4
    >>> input_tensor = np.random.rand(1, in_channels, 32, 32).astype(np.float32)  # (Batch, Channels, Height, Width)
    >>> input_tensor_torch = torch.from_numpy(input_tensor)
    >>> layer = Stem(in_channels, out_channels)
    >>> output_tensor = layer(input_tensor_torch)
    >>> output_tensor.shape
    torch.Size([1, 4, 15, 15])
    """

    def __init__(self, in_channels: int, out_channels: int) -> None:
        """
        Initializes the Stem layer.

        Parameters
        ----------
        in_channels : int
            The number of channels in the input tensor.
        out_channels : int
            The number of filters applied in the convolution operation.
        """
        super(Stem, self).__init__()

        self.conv_layer = nn.Conv2d(in_channels=in_channels,
                                    out_channels=out_channels,
                                    kernel_size=4,
                                    stride=2)

        self.activation_layer = nn.ReLU(inplace=True)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Stem layer.

        Parameters
        ----------
        inputs : torch.Tensor
            Input tensor of shape (batch_size, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch_size, out_channels, H_out, W_out), where
            H_out and W_out are reduced due to downsampling. The output is a
            feature map with extracted spatial representations.
        """
        conv1 = self.conv_layer(inputs)
        return self.activation_layer(conv1)



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 06 14:25:40 2017

@author: Zhenqin Wu
"""
import os
import time
import csv
import numpy as np
import deepchem
import pickle
from deepchem.molnet.run_benchmark_models import benchmark_classification, benchmark_regression
from deepchem.molnet.check_availability import CheckFeaturizer, CheckSplit
from deepchem.molnet.preset_hyper_parameters import hps


def run_benchmark(datasets,
                  model,
                  split=None,
                  metric=None,
                  direction=True,
                  featurizer=None,
                  n_features=0,
                  out_path='.',
                  hyper_parameters=None,
                  hyper_param_search=False,
                  max_iter=20,
                  search_range=2,
                  test=False,
                  reload=True,
                  seed=123):
    """
    Run benchmark test on designated datasets with deepchem(or user-defined) model

    Parameters
    ----------
    datasets: list of string
        choice of which datasets to use, should be: bace_c, bace_r, bbbp, chembl,
        clearance, clintox, delaney, hiv, hopv, kaggle, lipo, muv, nci, pcba,
        pdbbind, ppb, qm7, qm7b, qm8, qm9, sampl, sider, tox21, toxcast, uv, factors,
        kinase
    model: string or user-defined model stucture
        choice of which model to use, deepchem provides implementation of
        logistic regression, random forest, multitask network,
        bypass multitask network, irv, graph convolution;
        for user define model, it should include function: fit, evaluate
    split: string,  optional (default=None)
        choice of splitter function, None = using the default splitter
    metric: string, optional (default=None)
        choice of evaluation metrics, None = using the default metrics(AUC & R2)
    direction: bool, optional(default=True)
        Optimization direction when doing hyperparameter search
        Maximization(True) or minimization(False)
    featurizer: string or dc.feat.Featurizer,  optional (default=None)
        choice of featurization, None = using the default corresponding to model
        (string only applicable to deepchem models)
    n_features: int, optional(default=0)
        depending on featurizers, redefined when using deepchem featurizers,
        need to be specified for user-defined featurizers(if using deepchem models)
    out_path: string, optional(default='.')
        path of result file
    hyper_parameters: dict, optional (default=None)
        hyper parameters for designated model, None = use preset values
    hyper_param_search: bool, optional(default=False)
        whether to perform hyper parameter search, using gaussian process by default
    max_iter: int, optional(default=20)
        number of optimization trials
    search_range: int(float), optional(default=4)
        optimization on [initial values / search_range,
            initial values * search_range]
    test: boolean, optional(default=False)
        whether to evaluate on test set
    reload: boolean, optional(default=True)
        whether to save and reload featurized datasets
    """
    for dataset in datasets:
        if dataset in [
                'bace_c', 'bbbp', 'clintox', 'hiv', 'muv', 'pcba', 'pcba_146',
                'pcba_2475', 'sider', 'tox21', 'toxcast'
        ]:
            mode = 'classification'
            if metric is None:
                metric = [
                    deepchem.metrics.Metric(deepchem.metrics.roc_auc_score,
                                            np.mean),
                ]
        elif dataset in [
                'bace_r', 'chembl', 'clearance', 'delaney', 'hopv', 'kaggle',
                'lipo', 'nci', 'pdbbind', 'ppb', 'qm7', 'qm7b', 'qm8', 'qm9',
                'sampl', 'thermosol'
        ]:
            mode = 'regression'
            if metric is None:
                metric = [
                    deepchem.metrics.Metric(deepchem.metrics.pearson_r2_score,
                                            np.mean)
                ]
        else:
            raise ValueError('Dataset not supported')

        if featurizer is None and isinstance(model, str):
            # Assigning featurizer if not user defined
            pair = (dataset, model)
            if pair in CheckFeaturizer:
                featurizer = CheckFeaturizer[pair][0]
                n_features = CheckFeaturizer[pair][1]
            else:
                continue

        if split not in [None] + CheckSplit[dataset]:
            continue

        loading_functions = {
            'bace_c': deepchem.molnet.load_bace_classification,
            'bace_r': deepchem.molnet.load_bace_regression,
            'bbbp': deepchem.molnet.load_bbbp,
            'chembl': deepchem.molnet.load_chembl,
            'clearance': deepchem.molnet.load_clearance,
            'clintox': deepchem.molnet.load_clintox,
            'delaney': deepchem.molnet.load_delaney,
            'factors': deepchem.molnet.load_factors,
            'hiv': deepchem.molnet.load_hiv,
            'hopv': deepchem.molnet.load_hopv,
            'hppb': deepchem.molnet.load_hppb,
            'kaggle': deepchem.molnet.load_kaggle,
            'kinase': deepchem.molnet.load_kinase,
            'lipo': deepchem.molnet.load_lipo,
            'muv': deepchem.molnet.load_muv,
            'nci': deepchem.molnet.load_nci,
            'pcba': deepchem.molnet.load_pcba,
            'pdbbind': deepchem.molnet.load_pdbbind,
            'ppb': deepchem.molnet.load_ppb,
            'qm7': deepchem.molnet.load_qm7,
            'qm8': deepchem.molnet.load_qm8,
            'qm9': deepchem.molnet.load_qm9,
            'sampl': deepchem.molnet.load_sampl,
            'sider': deepchem.molnet.load_sider,
            'thermosol': deepchem.molnet.load_thermosol,
            'tox21': deepchem.molnet.load_tox21,
            'toxcast': deepchem.molnet.load_toxcast,
            'uv': deepchem.molnet.load_uv,
        }

        print('-------------------------------------')
        print('Benchmark on dataset: %s' % dataset)
        print('-------------------------------------')
        # loading datasets
        if split is not None:
            print('Splitting function: %s' % split)
            tasks, all_dataset, transformers = loading_functions[dataset](
                featurizer=featurizer, split=split, reload=reload)
        else:
            tasks, all_dataset, transformers = loading_functions[dataset](
                featurizer=featurizer, reload=reload)

        train_dataset, valid_dataset, test_dataset = all_dataset

        time_start_fitting = time.time()
        train_score = {}
        valid_score = {}
        test_score = {}

        if hyper_param_search:
            if hyper_parameters is None:
                hyper_parameters = hps[model]
            search_mode = deepchem.hyper.GaussianProcessHyperparamOpt(model)
            hyper_param_opt, _ = search_mode.hyperparam_search(
                hyper_parameters,
                train_dataset,
                valid_dataset,
                transformers,
                metric,
                direction=direction,
                n_features=n_features,
                n_tasks=len(tasks),
                max_iter=max_iter,
                search_range=search_range)
            hyper_parameters = hyper_param_opt
        if isinstance(model, str):
            if mode == 'classification':
                train_score, valid_score, test_score = benchmark_classification(
                    train_dataset,
                    valid_dataset,
                    test_dataset,
                    tasks,
                    transformers,
                    n_features,
                    metric,
                    model,
                    test=test,
                    hyper_parameters=hyper_parameters,
                    seed=seed)
            elif mode == 'regression':
                train_score, valid_score, test_score = benchmark_regression(
                    train_dataset,
                    valid_dataset,
                    test_dataset,
                    tasks,
                    transformers,
                    n_features,
                    metric,
                    model,
                    test=test,
                    hyper_parameters=hyper_parameters,
                    seed=seed)
        else:
            model.fit(train_dataset)
            train_score['user_defined'] = model.evaluate(
                train_dataset, metric, transformers)
            valid_score['user_defined'] = model.evaluate(
                valid_dataset, metric, transformers)
            if test:
                test_score['user_defined'] = model.evaluate(
                    test_dataset, metric, transformers)

        time_finish_fitting = time.time()

        with open(os.path.join(out_path, 'results.csv'), 'a') as f:
            writer = csv.writer(f)
            model_name = list(train_score.keys())[0]
            for i in train_score[model_name]:
                output_line = [
                    dataset,
                    str(split), mode, model_name, i, 'train',
                    train_score[model_name][i], 'valid',
                    valid_score[model_name][i]
                ]
                if test:
                    output_line.extend(['test', test_score[model_name][i]])
                output_line.extend([
                    'time_for_running', time_finish_fitting - time_start_fitting
                ])
                writer.writerow(output_line)
        if hyper_param_search:
            with open(os.path.join(out_path, dataset + model + '.pkl'),
                      'w') as f:
                pickle.dump(hyper_parameters, f)


#
# Note by @XericZephyr. Reason why I spun off this function:
#   1. Some model needs dataset information.
#   2. It offers us possibility to **cache** the dataset
#      if the featurizer runs very slow, e.g., GraphConv.
#   2+. The cache can even happen at Travis CI to accelerate
#       CI testing.
#
def load_dataset(dataset, featurizer, split='random'):
    """
    Load specific dataset for benchmark.

    Parameters
    ----------
    dataset: string
        choice of which datasets to use, should be: tox21, muv, sider,
        toxcast, pcba, delaney, factors, hiv, hopv, kaggle, kinase, nci,
        clintox, hiv, pcba_128, pcba_146, pdbbind, chembl, qm7, qm7b, qm9,
        sampl, uv
    featurizer: string or dc.feat.Featurizer.
        choice of featurization.
    split: string,  optional (default=None)
        choice of splitter function, None = using the default splitter
    """
    dataset_loading_functions = {
        'bace_c': deepchem.molnet.load_bace_classification,
        'bace_r': deepchem.molnet.load_bace_regression,
        'bbbp': deepchem.molnet.load_bbbp,
        'chembl': deepchem.molnet.load_chembl,
        'clearance': deepchem.molnet.load_clearance,
        'clintox': deepchem.molnet.load_clintox,
        'delaney': deepchem.molnet.load_delaney,
        'factors': deepchem.molnet.load_factors,
        'hiv': deepchem.molnet.load_hiv,
        'hopv': deepchem.molnet.load_hopv,
        'hppb': deepchem.molnet.load_hppb,
        'kaggle': deepchem.molnet.load_kaggle,
        'kinase': deepchem.molnet.load_kinase,
        'lipo': deepchem.molnet.load_lipo,
        'muv': deepchem.molnet.load_muv,
        'nci': deepchem.molnet.load_nci,
        'pcba': deepchem.molnet.load_pcba,
        'pcba_128': deepchem.molnet.load_pcba_128,
        'pcba_146': deepchem.molnet.load_pcba_146,
        'pcba_2475': deepchem.molnet.load_pcba_2475,
        'pdbbind': deepchem.molnet.load_pdbbind,
        'ppb': deepchem.molnet.load_ppb,
        'qm7': deepchem.molnet.load_qm7,
        'qm8': deepchem.molnet.load_qm8,
        'qm9': deepchem.molnet.load_qm9,
        'sampl': deepchem.molnet.load_sampl,
        'sider': deepchem.molnet.load_sider,
        'thermosol': deepchem.molnet.load_thermosol,
        'tox21': deepchem.molnet.load_tox21,
        'toxcast': deepchem.molnet.load_toxcast,
        'uv': deepchem.molnet.load_uv
    }
    print('-------------------------------------')
    print('Loading dataset: %s' % dataset)
    print('-------------------------------------')
    # loading datasets
    if split is not None:
        print('Splitting function: %s' % split)
    tasks, all_dataset, transformers = dataset_loading_functions[dataset](
        featurizer=featurizer, split=split)
    return tasks, all_dataset, transformers


def benchmark_model(model, all_dataset, transformers, metric, test=False):
    """
    Benchmark custom model.

    model: user-defined model stucture
        For user define model, it should include function: fit, evaluate.

    all_dataset: (train, test, val) data tuple.
        Returned by `load_dataset` function.

    transformers

    metric: string
        choice of evaluation metrics.
    """
    time_start_fitting = time.time()
    train_score = .0
    valid_score = .0
    test_score = .0

    train_dataset, valid_dataset, test_dataset = all_dataset

    model.fit(train_dataset)
    train_score = model.evaluate(train_dataset, metric, transformers)
    valid_score = model.evaluate(valid_dataset, metric, transformers)
    if test:
        test_score = model.evaluate(test_dataset, metric, transformers)

    time_finish_fitting = time.time()
    time_for_running = time_finish_fitting - time_start_fitting

    return train_score, valid_score, test_score, time_for_running



from rdkit import Chem
import numpy as np
from typing import List, Tuple
import re


def handle_hydrogen(smiles: str,
                    keep_h: bool = False,
                    add_h: bool = False) -> Chem.rdchem.Mol:
    """
    Builds an RDKit molecule from a SMILES string by conditionally
    handling addition or removal of hydrogens.

    This function is useful to process organic molecules properly.

    Examples
    --------
    >>> import deepchem as dc
    >>> mol = dc.utils.handle_hydrogen('C', keep_h=True, add_h=False)
    >>> mol.GetNumAtoms()
    1
    >>> mol = dc.utils.handle_hydrogen('C', keep_h=True, add_h=True)
    >>> mol.GetNumAtoms()
    5

    Parameters
    ----------
    smiles : str
        SMILES string.
    keep_h : bool, optional, default=False
        Whether to keep hydrogens in the molecule.
    add_h : bool, optional, default=False
        Whether to add hydrogens to the molecule.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        RDKit molecule object.

    """
    if keep_h:
        mol = Chem.MolFromSmiles(smiles, sanitize=False)
        Chem.SanitizeMol(mol,
                         sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL ^
                         Chem.SanitizeFlags.SANITIZE_ADJUSTHS)
    else:
        mol = Chem.MolFromSmiles(smiles)
    if add_h:
        mol = Chem.AddHs(mol)
    return mol


def make_polymer_mol(
    smiles: str,
    fragment_weights: List[float],
    keep_h: bool,
    add_h: bool,
) -> Chem.rdchem.Mol:
    """
    Builds an RDKit joined molecule from a SMILES string of monomer molecules.
    The weight of each monomer is stored as a metadata property of each atom.

    Examples
    --------
    >>> import deepchem as dc
    >>> mol = dc.utils.make_polymer_mol('C.C', [1, 1], True, False)
    >>> for atom in mol.GetAtoms():
    ...     print(atom.GetDoubleProp('w_frag'))
    1.0
    1.0

    Parameters
    ----------
    smiles : str
        SMILES string of monomer molecules joined by ".".
    fragment_weights : list
        List of weights for each monomer.
    keep_h : bool
        Whether to keep hydrogens in the molecule.
    add_h : bool
        Whether to add hydrogens to the molecule.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        RDKit polymer molecule object.
    """

    # check input is correct, we need the same number of fragments and their weights
    num_frags = len(smiles.split('.'))
    if len(fragment_weights) != num_frags:
        raise ValueError(
            f'number of input monomers/fragments ({num_frags}) does not match number of '
            f'input number of weights ({len(fragment_weights)})')

    # if it all looks good, we create one molecule object per fragment, add the weight as property
    # of each atom, and merge fragments into a single molecule object
    mols = []
    for s, w in zip(smiles.split('.'), fragment_weights):
        m = handle_hydrogen(s, keep_h, add_h)
        for a in m.GetAtoms():
            a.SetDoubleProp(
                'w_frag', float(w)
            )  # assinging metadata to store the weight fragment of the monomer
        mols.append(m)

    # combine all mols into single mol object
    mol = mols.pop(0)
    while len(mols) > 0:
        m2 = mols.pop(0)
        mol = Chem.CombineMols(mol, m2)

    return mol


def parse_polymer_rules(rules: List[str]) -> Tuple[List[tuple], float]:
    """
    This function extracts probabilty weight distribution details for bonds
    from string to list of tuples in following format.
    (start, end, weight_forward, weight_reverse)
    The start and end is in string. (As it will be mapped with open bond string)
    The weight_forward and weight_reverse in float
    It also returns the degree of polymerization of the polymer (DoP) as 1. + np.log10(DoP).
    The DoP can be mentioned at the end of the rules separated by "~".
    If DoP is not given, the function assumes DoP = 1.

    Examples
    --------
    >>> import deepchem as dc
    >>> polymer_info, degree_of_polymerization = dc.utils.parse_polymer_rules(
    ...     ['1-2:0.5:0.5'])
    >>> polymer_info
    [('1', '2', 0.5, 0.5)]
    >>> degree_of_polymerization
    1.0

    Parameters
    ----------
    rules : list[str]
        List of strings containing bond rules in the format
        "start-end:weight_forward:weight_reverse"
    Returns:
    --------
    polymer_info : list[tuple]
        List of tuples containing bond rules in the format
        (start, end, weight_forward, weight_reverse)
    degree_of_polymerization : float
        Degree of polymerization of the polymer.
    """

    polymer_info = []

    # check if deg of polymerization is provided
    if '~' in rules[-1]:
        Xn = float(rules[-1].split('~')[1])
        rules[-1] = rules[-1].split('~')[0]
    else:
        Xn = 1.

    for rule in rules:
        # handle edge case where we have no rules, and rule is empty string
        if rule == "":
            continue
        # QC of input string
        if len(rule.split(':')) != 3:
            raise ValueError(f'incorrect format for input information "{rule}"')
        if len(rule.split(":")[0].split("-")) != 2:
            raise ValueError(
                f'incorrect format for bond index mentioning "{rule}"')
        idx1, idx2 = rule.split(':')[0].split('-')
        w12 = float(rule.split(':')[1])  # weight for bond R_idx1 -> R_idx2
        w21 = float(rule.split(':')[2])  # weight for bond R_idx2 -> R_idx1
        polymer_info.append((idx1, idx2, w12, w21))

    return polymer_info, 1. + np.log10(Xn)


def tag_atoms_in_repeating_unit(
        mol: Chem.rdchem.RWMol) -> Tuple[Chem.rdchem.RWMol, dict]:
    """
    This function tags atoms that are part of the core units, as well as atoms
    serving to identify attachment points. In addition, create a map of bond
    types based on what bonds are connected to R groups in the input. The input molecules must be
    of `Chem.rdchem.RWMol` type to be editable.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> rw_mol = Chem.rdchem.RWMol(Chem.MolFromSmiles('[1*]CC.C[2*]'))
    >>> mol, _ = dc.utils.tag_atoms_in_repeating_unit(rw_mol)
    >>> mol.GetAtomWithIdx(0).GetBoolProp('core')
    False
    >>> mol.GetAtomWithIdx(1).GetBoolProp('core')
    True
    >>> mol.GetAtomWithIdx(3).GetProp('R')
    '2*'

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.RWMol
        RDKit read and write enabled molecule object.

    Returns
    -------
    rdkit.Chem.rdchem.RWMol
        RDKit read and write enabled molecule object.
    dict
        Map of R group to bond type.
    """

    atoms = [a for a in mol.GetAtoms()]
    neighbor_map = {}  # map R group to index of atom it is attached to
    r_bond_types = {}  # map R group to bond type

    # go through each atoms and: (i) get index of attachment atoms, (ii) tag all non-R atoms
    for atom in atoms:
        # if R atom
        if '*' in atom.GetSmarts():
            # get index of atom it is attached to
            neighbors = atom.GetNeighbors()
            assert len(neighbors) == 1
            neighbor_idx = neighbors[0].GetIdx()
            r_tag = atom.GetSmarts().strip('[]').replace(':', '')  # *1, *2, ...
            neighbor_map[r_tag] = neighbor_idx
            # tag it as non-core atom
            atom.SetBoolProp('core', False)
            # create a map R --> bond type
            bond = mol.GetBondBetweenAtoms(atom.GetIdx(), neighbor_idx)
            r_bond_types[r_tag] = bond.GetBondType()
        # if not R atom
        else:
            # tag it as core atom
            atom.SetBoolProp('core', True)

    # use the map created to tag attachment atoms
    for atom in atoms:
        if atom.GetIdx() in neighbor_map.values():
            r_tags = [k for k, v in neighbor_map.items() if v == atom.GetIdx()]
            atom.SetProp('R', ''.join(r_tags))
        else:
            atom.SetProp('R', '')

    return mol, r_bond_types


def onek_encoding_unk(value: int, choices: list) -> list:
    """
    This function generates the vector for a value as one hot encoded list.
    If there is an unknown value, it will be encoded as 1 in the last index

    Examples
    --------
    >>> import deepchem as dc
    >>> dc.utils.onek_encoding_unk(1, [1, 2, 3])
    [1, 0, 0, 0]
    >>> dc.utils.onek_encoding_unk(69, [1, 2, 3])
    [0, 0, 0, 1]

    Parameters
    ----------
    value : int
        Value to be encoded.
    choices : list
        List of choices.

    Returns
    -------
    list
        One hot encoded vector.
    """

    encoding = [0] * (len(choices) + 1)
    index = choices.index(value) if value in choices else -1
    encoding[index] = 1

    return encoding


def remove_wildcard_atoms(rwmol: Chem.rdchem.RWMol) -> Chem.rdchem.RWMol:
    """
    This function removes the connection virtual atoms for open bonds in a molecule.
    This is necessary for molecules with wildcard notations.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles("[*]CC")
    >>> rwmol = Chem.RWMol(mol)
    >>> rwmol = dc.utils.remove_wildcard_atoms(rwmol)
    >>> Chem.MolToSmiles(rwmol)
    'CC'

    Parameters
    ----------
    rwmol : rdkit.Chem.rdchem.RWMol
        Read and Writable RDKit molecule object.

    Returns
    -------
    rdkit.Chem.rdchem.RWMol
        Read and writable RDKit molecule object.
    """
    indices = [a.GetIdx() for a in rwmol.GetAtoms() if '*' in a.GetSmarts()]
    while len(indices) > 0:
        rwmol.RemoveAtom(indices[0])  # removing the wildcard atom
        indices = [a.GetIdx() for a in rwmol.GetAtoms() if '*' in a.GetSmarts()]
    Chem.SanitizeMol(rwmol, Chem.SanitizeFlags.SANITIZE_ALL)
    return rwmol


class PolyWDGStringValidator():
    """
    Class for validating the string format of weighted directed graph
    data.This class provides methods to validate the format of a
    datapoint string. This is a specific string format that is used
    for storing weighted directed polymer data in a parsable format.

    The format is as follows:
        [monomer1].[monomer2]|[fraction_of_monomer1]|[fraction_of_monomer2]|<[polymer_rule1]<[polymer_rule2]
    The polymer rule has an own format in it. Which is as follows:
        [[atom_index1]-[atom_index2]]:[fraction_of_bond_between_atom1_to_atom2]:[fraction_of_bond_between_atom2_to_atom1]

    This format is explicitly used for formatting the input for
    Weighted Directed Message Passing Neural Networks (wD-MPNN).
    The input format holds a SMART notation and regular expression
    formatting to keep molecular data with corresponding bonds and
    weights. Irrespective of this explicit usecase, the formatting
    can allow featurization of same data for other graph based neural
    networks.

    The validate method validates the proper formatting for monomer
    molecules, proper value of the fractions and valid atom indicies
    and corresponding weights in the polymer rules.

    Example
    -------
    >>> from deepchem.utils import PolyWDGStringValidator
    >>> validator = PolyWDGStringValidator()
    >>> validator.validate("[1*]C.C[2*]|0.5|0.5|<1-2:0.5:0.5")
    True

    References
    ----------
    .. [1] Aldeghi, Matteo, and Connor W. Coley. "A graph representation of molecular
        ensembles for polymer property prediction." Chemical Science 13.35 (2022): 10486-10498.
    """

    @staticmethod
    def get_parsed_vals(datapoint: str) -> Tuple[str, list, str]:
        """
        This static method parses the datapoint string into 3 parts:
        1. Monomer molecules
        2. Fragments
        3. Polymer rules

        Parameters
        ----------
        datapoint : str
            The datapoint string to parse

        Returns
        -------
        Tuple[str, list, str]
            A tuple containing the 3 parts of the datapoint string
        """
        base_parsed = datapoint.split("|")
        if len(base_parsed) < 3:
            raise ValueError(
                f"Invalid datapoint format: At least 3 splits should be there but found {len(base_parsed)} no. of splits"
            )
        monomer_mols = base_parsed[0]
        polymer_rules = base_parsed[-1]
        fragments = base_parsed[1:-1]
        return monomer_mols, fragments, polymer_rules

    @staticmethod
    def get_polymer_rules(rules_str: str) -> List[str]:
        """
        This static method parses the polymer rules string into a list of rules.

        Parameters
        ----------
        rules_str : str
            The polymer rules string to parse

        Returns
        -------
        List[str]
            A list containing the parsed rule strings
        """
        if len(rules_str.split("<")) == 1:
            raise ValueError(
                "Invalid rules string: The rule string must contain '<' as a separator for rules !"
            )
        return rules_str.split("<")[1:]

    def _validate_fragments(self, datapoint: str):
        """
        This method validate the number of fragments match
        the number of monomers.

        Parameters
        ----------
        datapoint : str
            The datapoint string to validate

        Raises
        ------
        ValueError
            If the number of fragments does not match the number of monomers
        """
        monomer_mols, fragments, _ = self.get_parsed_vals(datapoint)
        if len(fragments) != len(monomer_mols.split(".")):
            raise ValueError(
                f"Number of fragments and number of monomers should match. Mismatch -> No. of Fragments {len(fragments)} , No. of Monomers{len(monomer_mols.split('.'))}"
            )

    def _get_all_wildcards(self, text: str) -> List[str]:
        """
        This method returns all the wildcards present in the given string
        representation by using regular expression to detect digits after
        '*'.

        Parameters
        ----------
        text : str

        Returns
        -------
        List[str]
            A list of all wildcards present in the text
        """
        matches = re.findall(r"\d+(?=\*)", text)
        return matches

    def _validate_wildcards(self, datapoint: str):
        """
        This method validates the presence of wildcards in the polymer
        molecules string and ensures that the sequence of the wildcard
        notation is proper.

        Parameters
        ----------
        datapoint : str
            The datapoint string to validate

        Raises
        ------
        ValueError
            If the wildcards are not present in the sequce the maximum
            wildcard value, ValueError is raised.
        """
        monomer_mols, _, _ = self.get_parsed_vals(datapoint)
        max_wildcard = max(
            [int(x) for x in self._get_all_wildcards(monomer_mols)])
        for wildcard in range(1, max_wildcard + 1):
            if str(wildcard) + "*" not in monomer_mols:
                raise ValueError(
                    f"Invalid wildcard format: The wildcard {wildcard} is not present in the monomer molecules string  as per the sequence of the maximum {max_wildcard}!"
                )

    def _validate_polymer_rules(self, datapoint: str):
        """
        This method validates the format of the polymer rules string
        by checking for the presence of the '-' separator between the
        atom indexes, the correct number of splits in the rule string,
        and the validity of the atom indexes present in the monomer
        SMILES. It also checks if the atom indexes are in the correct
        correct count for a valid bond formation.

        Parameters
        ----------
        datapoint : str
            The datapoint string to validate

        Raises
        ------
        ValueError
            If the polymer rules string is invalid, ValueError is raised
            with appropriate error messages
        """
        monomer_mols, _, polymer_rules = self.get_parsed_vals(datapoint)
        polymer_rule_list = self.get_polymer_rules(polymer_rules)
        for rules in polymer_rule_list:
            splits = rules.split(":")
            if len(splits) != 3:
                raise ValueError(
                    f"Invalid polymer rule format: The rule must contain exactly 3 splits ! but found {len(splits)} splits"
                )
            if "-" not in splits[0]:
                raise ValueError(
                    f"Invalid polymer rule format: The bond string between two wildcard index must be seprated by '-', got invalid data {splits[0]}"
                )
            elif len(splits[0].split("-")) != 2 and any(
                    elem == "" for elem in splits[0].split("-")):
                raise ValueError(
                    f"Invalid polymer rule format: The first split must contain exactly 2 splits to depict connection between atom indexes! but found {len(splits[0].split('-'))} splits"
                )
            else:
                for wild_card_index in splits[0].split("-"):
                    if not wild_card_index.isdigit():
                        raise ValueError(
                            f"Invalid polymer rule format: The first split must contain only digits! but found {wild_card_index}"
                        )
                    if wild_card_index not in monomer_mols:
                        raise ValueError(
                            f"Invalid polymer rule format: The first split must contain only valid wild card indexes! but found {wild_card_index} which is not in {monomer_mols}"
                        )

    def validate(self, datapoint: str):
        """
        This method validates the string format of weighted
        directed graph data. To validate the string format
        it checks for following conditions:

        1. The number of fragments and the number of monomer
           molecules should match.
        2. The wild card indexes should be present in the monomer
           molecules string and should be in the correct sequence.
        3. The polymer rules should be in the correct format.
        4. The atom indexes in the polymer rules should be valid
           and present in the monomer molecules string.

        It raises ValueError if the string format is invalid.

        Parameters
        ----------
        datapoint : str
            The datapoint string to validate

        Returns
        -------
        bool
            True if the string format is valid, None otherwise
            (Error will be raised otherwise)
        """
        self._validate_fragments(datapoint)
        self._validate_wildcards(datapoint)
        self._validate_polymer_rules(datapoint)
        return True



import torch.nn as nn
from torch import Tensor
from transformers import PreTrainedTokenizer, BertTokenizer
from torch.nn.utils.rnn import pad_sequence
from torch.nn.modules.loss import _Loss
from deepchem.models.losses import Loss
from deepchem.models.torch_models import TorchModel
from deepchem.models.optimizers import Optimizer, Adam
from deepchem.data import Dataset, NumpyDataset
import torch
from typing import Iterable, Optional, Union, Tuple, List, Any, Dict, Sequence


class LSTMNeuralNet(nn.Module):
    """
    An LSTM-based neural network for token generation or classification tasks.
    This class contains an embedding layer, an LSTM layer, and a fully connected
    layer to translate hidden states into vocabulary tokens.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.lstm_generator_models import LSTMNeuralNet
    >>> model = LSTMNeuralNet(vocab_size=12, embedding_dim=8, hidden_dim=4, num_layers=1)
    >>> x = torch.randint(0, 12, (2, 10))
    >>> output = model(x)
    """

    def __init__(self,
                 vocab_size: int,
                 embedding_dim: int,
                 hidden_dim: int,
                 num_layers: int = 1) -> None:
        """
        Initialize the LSTMNeuralNet.

        Parameters
        ----------
        vocab_size: int
            Size of the vocabulary.
        embedding_dim: int
            Dimension of the embedding vector.
        hidden_dim: int
            Number of hidden units in each LSTM layer.
        num_layers: int
            Number of LSTM layers in the network.
        """
        super(LSTMNeuralNet, self).__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        self.rnn = nn.LSTM(embedding_dim,
                           hidden_dim,
                           num_layers,
                           batch_first=True)
        self.fc = nn.Linear(hidden_dim, vocab_size)

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the LSTMNeuralNet.

        Parameters
        ----------
        x: Tensor
            Input tensor of token ids.

        Returns
        -------
        Tensor
            Raw logits of shape [batch_size, sequence_length, vocab_size].
        """
        embedded = self.embedding(x)
        output, _ = self.rnn(embedded)
        output = self.fc(output)
        return output


class LSTMGenerator(TorchModel):
    """LSTM Generator

    This class implements an LSTM-based [1]_ generator for token generation tasks.
    The generator is trained on a list of sequences and can be used to generate
    new sequences of tokens. The model is implemented using PyTorch and can be useful
    to generate SMILES, PSMILES and Weighted Graph strings for generation tasks.
    The generator is used in our research paper "Open-source Polymer Generative Pipeline" [2]_ to generate
    hypothetical polymers using PSMILES and Weighted Directed Graph representations.

    References
    ----------
    .. [1] Staudemeyer, Ralf C., and Eric Rothstein Morris. "Understanding LSTM
        --a tutorial into long short-term memory recurrent neural networks."
        arXiv preprint arXiv:1909.09586 (2019).
    .. [2] Mohanty, Debasish, et al. "Open-source Polymer Generative Pipeline."
        arXiv preprint arXiv:2412.08658 (2024).

    Examples
    --------
    >>> from deepchem.models.torch_models import LSTMGenerator
    >>> from deepchem.data import NumpyDataset
    >>> generator = LSTMGenerator()
    >>> dataset = NumpyDataset(["CCC"])
    >>> loss = generator.fit(dataset, nb_epoch=1)
    >>> new_sequences = generator.sample(num_gen=10, max_len=10, temperature=1.0)
    """

    def __init__(
            self,
            loss: Union[Loss, _Loss] = nn.CrossEntropyLoss(),
            # There is a problem with deepchem's SparseSoftmaxCrossEntropy.
            # The reduction='none' kwarg is not allowing backpropagation gradients
            batch_size: int = 8,
            embedding_dim: int = 128,
            hidden_dim: int = 256,
            num_layers: int = 2,
            tokenizer: Optional[PreTrainedTokenizer] = None,
            learning_rate: Union[float] = 0.001,
            optimizer: Optional[Optimizer] = None,
            model_dir: str = "lstm_generator_model",
            device: Optional[torch.device] = None) -> None:
        """
        Initializes the LSTMGenerator model.

        Parameters
        ----------
        loss: Loss, default nn.CrossEntropyLoss
            Loss function to use for training the model.
        batch_size: int, default 8
            Batch size to use during training.
        embedding_dim: int, default 128
            Dimension of the embedding vector.
        hidden_dim: int, default 256
            Number of hidden units in each LSTM layer.
        num_layers: int, default 2
            Number of LSTM layers in the network.
        tokenizer: PreTrainedTokenizer, default None
            Tokenizer to use for tokenizing input sequences.
        learning_rate: float, default 0.001
            Learning rate to use during training.
        optimizer: Optimizer, default None
            Optimizer to use for training the model. If None, Adam optimizer is used.
        model_dir: str, default "lstm_generator_model"
            Directory to save the model files.
        device: torch.device, default None
            Device to use for training the model.
        """

        self.tokenizer = BertTokenizer.from_pretrained(
            'bert-base-cased') if tokenizer is None else tokenizer
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        model = LSTMNeuralNet(vocab_size=self.tokenizer.vocab_size,
                              embedding_dim=embedding_dim,
                              hidden_dim=hidden_dim,
                              num_layers=num_layers)
        optimizer = optimizer if optimizer else Adam(
            learning_rate=learning_rate)
        if device is None:
            device = torch.device(
                'cuda' if torch.cuda.is_available() else 'cpu')
        model = model.to(device)
        super(LSTMGenerator, self).__init__(model=model,
                                            loss=loss,
                                            batch_size=batch_size,
                                            model_dir=model_dir,
                                            optimizer=optimizer,
                                            device=device,
                                            learning_rate=learning_rate)

    def default_generator(self,
                          dataset: Dataset,
                          epochs: int = 1,
                          *args,
                          **kwargs) -> Iterable[Tuple[List, List, List]]:
        """
        Generates a default generator for the input sequences.

        Parameters
        ----------
        dataset: Dataset
            Dataset of input sequences to tokenize.
        epochs: int, default 1
            Number of epochs to train the model.

        Returns
        -------
        Iterable
            Generator that yields input, target and zero weight
            as a tuple of tensors for training the model.
        """
        sequences = []
        for (X_, _, _, _) in dataset.iterbatches(batch_size=self.batch_size):
            # Tokenize the sequence and add special tokens
            for sequence in list(X_):
                tokens = self.tokenizer.encode(sequence,
                                               add_special_tokens=True)
                sequences.append(tokens)

        # Convert the list of sequences into tensors and pad them
        padded_sequences = pad_sequence(
            [torch.tensor(seq) for seq in sequences
            ],  # Convert each sequence to a tensor
            batch_first=True,
            padding_value=self.tokenizer.
            pad_token_id  # Use BERT's padding token ID
        )

        for epoch in range(epochs):
            for i in range(0, len(padded_sequences), self.batch_size):
                batch = padded_sequences[
                    i:min(i + self.batch_size, len(padded_sequences))]
                inputs = batch[:, :-1]
                targets = batch[:, 1:]
                weights = torch.ones(inputs.shape[0], device=self.device)
                yield ([inputs], [targets], [weights])

    def fit(self,
            dataset: Dataset,
            nb_epoch: int = 1,
            max_checkpoints_to_keep: int = 5,
            checkpoint_interval: int = 1000,
            *args,
            **kwargs) -> float:
        """
        Fits the model on the input sequences.

        Parameters
        ----------
        dataset: Dataset
            Dataset of input sequences to train the model.
        nb_epoch: int, default 1
            Number of epochs to train the model.
        max_checkpoints_to_keep: int, default 5
            Maximum number of checkpoints to keep.
        checkpoint_interval: int, default 1000
            Interval at which to save model checkpoints.

        Returns
        -------
        float
            Average loss of most recent checkpoint
        """
        return self.fit_generator(
            self.default_generator(dataset=dataset, epochs=nb_epoch),
            checkpoint_interval=checkpoint_interval,
            max_checkpoints_to_keep=max_checkpoints_to_keep)

    def fit_generator(self,
                      generator: Iterable,
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      *args,
                      **kwargs) -> float:
        """
        Fits the model on the input sequences using a generator.

        Parameters
        ----------
        generator: Iterable
            Generator that yields input, target and zero weight tensors for training the model.
        max_checkpoints_to_keep: int, default 5
            Maximum number of checkpoints to keep.
        checkpoint_interval: int, default 1000
            Interval at which to save model checkpoints.
        Returns
        -------
        float
            Average loss of most recent checkpoint
        """
        loss_float = 0.0
        current_step = 0
        optimizer = self.optimizer._create_pytorch_optimizer(
            self.model.parameters())
        for inputs, targets, _ in generator:
            input_val = inputs[0]
            input_val = input_val.to(self.device)
            target_val = targets[0]
            target_val = target_val.to(self.device)
            output = self.model(input_val)
            if isinstance(self.loss, Loss):
                # Convert DeepChem Loss to a PyTorch callable
                loss_fn = self.loss._create_pytorch_loss()
                loss_val = loss_fn(
                    output.reshape(-1, self.tokenizer.vocab_size),
                    target_val.reshape(-1))
            elif isinstance(self.loss, _Loss):
                # Already a PyTorch loss
                loss_val = self.loss(
                    output.reshape(-1, self.tokenizer.vocab_size),
                    target_val.reshape(-1))
            else:
                raise ValueError("Invalid loss type")

            # Backpropagation and optimize
            optimizer.zero_grad()
            loss_val.backward()
            optimizer.step()

            if checkpoint_interval > 0 and current_step % checkpoint_interval == checkpoint_interval - 1:
                self.save_checkpoint(max_checkpoints_to_keep)
            loss_float = loss_val.item()
            current_step += 1

        return float(loss_float)

    def restore(self,
                checkpoint: Optional[str] = None,
                model_dir: Optional[str] = None,
                strict: Optional[bool] = True) -> None:
        """Restore the values of all variables from a checkpoint file.

        Parameters
        ----------
        checkpoint: str
            the path to the checkpoint file to load.  If this is None, the most recent
            checkpoint will be chosen automatically.  Call get_checkpoints() to get a
            list of all available checkpoints.
        model_dir: str, default None
            Directory to restore checkpoint from. If None, use self.model_dir.  If
            checkpoint is not None, this is ignored.
        strict: bool, default True
            Whether or not to strictly enforce that the keys in checkpoint match
            the keys returned by this model's get_variable_scope() method.
        """
        self._ensure_built()
        if checkpoint is None:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            if len(checkpoints) == 0:
                raise ValueError('No checkpoint found')
            checkpoint = checkpoints[0]
        data = torch.load(checkpoint, map_location=self.device)
        self.model = LSTMNeuralNet(self.tokenizer.vocab_size,
                                   self.embedding_dim, self.hidden_dim,
                                   self.num_layers)
        if strict is not None:
            self.model.load_state_dict(data['model_state_dict'], strict=strict)
            return
        self.model.load_state_dict(data['model_state_dict'])

    def load_from_pretrained(self,
                             source_model: TorchModel,
                             assignment_map: Optional[Dict[Any, Any]] = None,
                             value_map: Optional[Dict[Any, Any]] = None,
                             checkpoint: Optional[str] = None,
                             model_dir: Optional[str] = None,
                             include_top: bool = True,
                             inputs: Optional[Sequence[Any]] = None,
                             **kwargs: Any) -> None:
        """
        Load model from a pretrained checkpoint.

        Parameters
        ----------
        source_model: TorchModel
            The source model to load from.
        assignment_map: dict, optional
            Parameter assignment (unused here).
        value_map: dict, optional
            Parameter assignment (unused here).
        checkpoint: str, optional
            Checkpoint file path.
        model_dir: str, optional
            Directory from which to restore the checkpoint.
        include_top: bool
            Whether to load the top layers (unused here).
        inputs: Sequence[Any], optional
            Input placeholders (unused here).
        """
        self._ensure_built()
        self.restore(model_dir=model_dir, checkpoint=checkpoint)
        self.model = self.model.to(self.device)

    def _predict(self, generator: Iterable[Tuple[Any, Any, Any]], *args: Any,
                 **kwargs: Any):
        """
        Predict the next token in the sequence.

        Parameters
        ----------
        generator: Iterable
            Input tensor of token ids.
        temperature: float, default 1.0
            (passed as keyword argument)
            Temperature to use for sampling.
        """
        if "temperature" in kwargs:
            temperature = kwargs["temperature"]
        else:
            temperature = 1.0
        for gen in generator:
            input_tensors, _, _ = gen
        input_tensor = torch.tensor(input_tensors)
        input_tensor = input_tensor.to(self.device)
        output = self.model(input_tensor)
        logits = output[:, -1, :] / temperature
        prbos = torch.softmax(logits, dim=-1)
        predicted_tokens = torch.multinomial(prbos, num_samples=1).item()
        return predicted_tokens

    def _single_sample(self,
                       max_len: int = 600,
                       temperature: float = 1.0) -> str:
        """
        Generate a sequence by single sampling.

        Parameters
        ----------
        max_len: int, default 600
            Maximum length of the generated sequence.
        temperature: float, default 1.0
            Temperature to use for sampling.

        Returns
        -------
        str
            Detokenized Generated sequence as str.
        """
        generated_sequence: list = [self.tokenizer.cls_token_id]
        for _ in range(max_len):
            input_tensor = torch.tensor(generated_sequence).unsqueeze(0)
            dataset = NumpyDataset(input_tensor)
            output_tensors: list = []

            def get_generator():
                for (X_, _, _,
                     _) in dataset.iterbatches(batch_size=self.batch_size):
                    yield X_, _, _

            output_tensor = self._predict(get_generator(),
                                          temperature=temperature)
            generated_sequence.append(output_tensor)
            if len(output_tensors) > 0:
                if output_tensors[0] == self.tokenizer.sep_token_id:
                    break
        generated_sequence_str: str = self.tokenizer.decode(
            generated_sequence, skip_special_tokens=True)
        generated_sequence_str = generated_sequence_str.replace(" ", "")
        return generated_sequence_str

    def sample(self,
               num_gen: int = 100,
               max_len: int = 600,
               temperature: float = 1.0) -> List[str]:
        """
        Generate sequences by repeated sampling.
        Parameters
        ----------
        num_gen: int, default 100
            Number of sequences to generate.
        max_len: int, default 600
            Maximum length of the generated sequence.
        temperature: float, default 1.0
            Temperature to use for sampling.
        Returns
        -------
        List[str]
            List of generated sequences as strings.
        """
        self.model.eval()
        results = []
        for i in range(num_gen):
            generated_sequence = self._single_sample(max_len=max_len,
                                                     temperature=temperature)
            results.append(generated_sequence)
        return results



import torch
import numpy as np
from typing import List, Optional, Tuple
from deepchem.utils.dft_utils import Lattice, BaseGrid, LebedevGrid


class BeckeGrid(BaseGrid):
    """
    Using Becke's scheme to construct the 3D grid consists of multiple 3D grids
    centered on each atom.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LebedevGrid, BeckeGrid, RadialGrid
    >>> grid = RadialGrid(100, grid_integrator="chebyshev", grid_transform="logm3")
    >>> atomgrid = [LebedevGrid(grid, 3), LebedevGrid(grid, 3)]
    >>> atompos = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
    >>> grid = BeckeGrid(atomgrid, atompos)
    >>> grid.get_rgrid().shape
    torch.Size([1200, 3])
    >>> grid.get_dvolume().shape
    torch.Size([1200])

    """

    def __init__(self,
                 atomgrid: List[LebedevGrid],
                 atompos: torch.Tensor,
                 atomradii: Optional[torch.Tensor] = None,
                 ratom_adjust: str = "becke") -> None:
        """Initialize the Becke grid.

        Parameters
        ----------
        atomgrid: List[LebedevGrid]
            List of Lebedev grids for each atom. (natoms)
        atompos: torch.Tensor
            Position of each atom. (natoms, ndim)
        atomradii: Optional[torch.Tensor] (default None)
            Radii of each atom. (natoms,) or None
        ratom_adjust: str (default "becke")
            Adjustment method for the atom radii. Available: ['becke', 'treutler']

        """
        assert atompos.shape[0] == len(atomgrid), \
            "The lengths of atomgrid and atompos must be the same"
        assert len(atomgrid) > 0
        self._dtype = atomgrid[0].dtype
        self._device = atomgrid[0].device

        # construct the grid points positions, weights, and the index of grid corresponding to each atom
        rgrids, self._rgrid, dvol_atoms = _construct_rgrids(atomgrid, atompos)

        # calculate the integration weights
        weights_atoms = _get_atom_weights(rgrids,
                                          atompos,
                                          atomradii=atomradii,
                                          ratom_adjust=ratom_adjust)  # (ngrid,)
        self._dvolume = dvol_atoms * weights_atoms

    @property
    def dtype(self):
        """Return the data type of the grid.

        Returns
        -------
        torch.dtype
            Data type of the grid.

        """
        return self._dtype

    @property
    def device(self):
        """device of the grid points

        Returns
        -------
        torch.device
            device of the grid points

        """
        return self._device

    @property
    def coord_type(self):
        """type of the coordinate returned in get_rgrid.

        Returns
        -------
        str
            type of the coordinate returned in get_rgrid. It can be 'cartesian'
            or 'spherical'.

        """
        return "cart"

    def get_dvolume(self):
        """Obtain the torch.tensor containing the dV elements for the integration.

        Returns
        -------
        torch.tensor (*BG, ngrid)
            The dV elements for the integration. *BG is the length of the BaseGrid.

        """
        return self._dvolume

    def get_rgrid(self) -> torch.Tensor:
        """
        Returns the grid points position in the specified coordinate in
        self.coord_type.

        Returns
        -------
        torch.tensor (*BG, ngrid, ndim)
            The grid points position. *BG is the length of the BaseGrid.

        """
        return self._rgrid

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        Return a list with the parameter names corresponding to the given method
        (methodname)

        Returns
        -------
        List[str]
            List of parameter names of methodname

        """
        if methodname == "get_rgrid":
            return [prefix + "_rgrid"]
        elif methodname == "get_dvolume":
            return [prefix + "_dvolume"]
        else:
            raise KeyError("Invalid methodname: %s" % methodname)


class PBCBeckeGrid(BaseGrid):
    """
    Use Becke's scheme to construct the 3D grid in a periodic cell. It is similar
    to non-pbc BeckeGrid, but in this case, only grid points inside the lattice
    are considered, and atoms corresponds to each grid points are involved in
    calculating the weights.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LebedevGrid, PBCBeckeGrid, RadialGrid, Lattice
    >>> grid = RadialGrid(100, grid_integrator="chebyshev", grid_transform="logm3")
    >>> atomgrid = [LebedevGrid(grid, 3), LebedevGrid(grid, 3)]
    >>> atompos = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=torch.float64)
    >>> lattice = Lattice(a)
    >>> grid = PBCBeckeGrid(atomgrid, atompos, lattice)
    >>> grid.get_rgrid().shape
    torch.Size([720, 3])
    >>> grid.get_dvolume().shape
    torch.Size([720])

    """

    def __init__(self,
                 atomgrid: List[LebedevGrid],
                 atompos: torch.Tensor,
                 lattice: Lattice,
                 ratom_adjust: str = "becke"):
        """Initialize the PBCBecke grid.

        Parameters
        ----------
        atomgrid: List[LebedevGrid]
            List of Lebedev grids for each atom. (natoms)
        atompos: torch.Tensor
            Position of each atom. (natoms, ndim)
        lattice: Lattice
            Lattice object for the periodic cell.
        ratom_adjust: str (default "becke")
            Adjustment method for the atom radii. Available: ['becke', 'treutler']

        """
        assert atompos.shape[0] == len(atomgrid), \
            "The lengths of atomgrid and atompos must be the same"
        assert len(atomgrid) > 0
        self._dtype = atomgrid[0].dtype
        self._device = atomgrid[0].device

        # get the normalized coordinates
        a = lattice.lattice_vectors()  # (nlvec=ndim, ndim)
        b = lattice.recip_vectors() / (
            2 * np.pi)  # (ndim, ndim) just the inverse of lattice vector.T

        new_atompos_lst: List[torch.Tensor] = []
        new_rgrids: List[torch.Tensor] = []
        new_dvols: List[torch.Tensor] = []
        for ia, atomgr in enumerate(atomgrid):
            atpos = atompos[ia]
            rgrid = atomgr.get_rgrid() + atpos  # (natgrid, ndim)
            dvols = atomgr.get_dvolume()  # (natgrid)

            # ugrid is the normalized coordinate
            ugrid = torch.einsum("cd,gd->gc", b, rgrid)  # (natgrid, ndim)

            # get the shift required to make the grid point inside the lattice
            ns = -ugrid.floor().to(
                torch.int
            )  # (natgrid, ndim) # ratoms + ns @ a will be the new atompos

            # ns_unique: (nunique, ndim), ns_unique_idx: (natgrid,), ns_count: (nunique)
            ns_unique, ns_unique_idx, ns_count = torch.unique(
                ns, dim=0, return_inverse=True, return_counts=True)

            # ignoring the shifts with only not more than 8 points (following pyscf)
            significant_uniq_idx = ns_count > 8  # (nunique)
            significant_idx = significant_uniq_idx[ns_unique_idx]  # (natgrid,)
            ns_unique = ns_unique[significant_uniq_idx, :]  # (nunique2, ndim)
            ls_unique = torch.matmul(ns_unique.to(a.dtype),
                                     a)  # (nunique2, ndim)

            # flag the unaccepted points with -1
            flag = -1
            ns_unique_idx[~significant_idx] = flag  # (natgrid)

            # get the coordinate inside the lattice
            ls = torch.matmul(ns.to(a.dtype), a)
            rg = rgrid + ls  # (natgrid, ndim)

            # get the new atom pos
            new_atpos = atompos[ia] + ls_unique  # (nunique2, ndim)
            new_atompos_lst.append(new_atpos)

            # separate the grid points that corresponds to the different atoms
            for idx in torch.unique(ns_unique_idx):
                if idx == flag:
                    continue
                at_idx = ns_unique_idx == idx
                new_rgrids.append(rg[at_idx, :])  # list of (natgrid2, ndim)
                new_dvols.append(dvols[at_idx])  # list of (natgrid2,)

        self._rgrid = torch.cat(new_rgrids, dim=0)  # (ngrid, ndim)
        dvol_atoms = torch.cat(new_dvols, dim=0)  # (ngrid)
        new_atompos = torch.cat(new_atompos_lst, dim=0)  # (nnewatoms, ndim)
        watoms = _get_atom_weights(new_rgrids,
                                   new_atompos,
                                   ratom_adjust=ratom_adjust)  # (ngrid,)
        self._dvolume = dvol_atoms * watoms

    @property
    def dtype(self):
        """Return the data type of the grid.

        Returns
        -------
        torch.dtype
            Data type of the grid.

        """
        return self._dtype

    @property
    def device(self):
        """device of the grid points

        Returns
        -------
        torch.device
            device of the grid points

        """
        return self._device

    @property
    def coord_type(self):
        """type of the coordinate returned in get_rgrid.

        Returns
        -------
        str
            type of the coordinate returned in get_rgrid. It can be 'cartesian'
            or 'spherical'.

        """
        return "cart"

    def get_dvolume(self):
        """Obtain the torch.tensor containing the dV elements for the integration.

        Returns
        -------
        torch.tensor (*BG, ngrid)
            The dV elements for the integration. *BG is the length of the BaseGrid.

        """
        return self._dvolume

    def get_rgrid(self) -> torch.Tensor:
        """
        Returns the grid points position in the specified coordinate in
        self.coord_type.

        Returns
        -------
        torch.tensor (*BG, ngrid, ndim)
            The grid points position. *BG is the length of the BaseGrid.

        """
        return self._rgrid

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        Return a list with the parameter names corresponding to the given method
        (methodname)

        Returns
        -------
        List[str]
            List of parameter names of methodname

        """
        if methodname == "get_rgrid":
            return [prefix + "_rgrid"]
        elif methodname == "get_dvolume":
            return [prefix + "_dvolume"]
        else:
            raise KeyError("Invalid methodname: %s" % methodname)


def _construct_rgrids(atomgrid: List[LebedevGrid], atompos: torch.Tensor) \
        -> Tuple[List[torch.Tensor], torch.Tensor, torch.Tensor]:
    """Construct the grid positions in a 2D tensor, the weights per isolated atom

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LebedevGrid, RadialGrid
    >>> from deepchem.utils.dft_utils.grid.multiatoms_grid import _construct_rgrids
    >>> grid = RadialGrid(100, grid_integrator="chebyshev", grid_transform="logm3")
    >>> atomgrid = [LebedevGrid(grid, 3), LebedevGrid(grid, 3)]
    >>> atompos = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
    >>> allpos_lst, rgrid, dvol_atoms = _construct_rgrids(atomgrid, atompos)
    >>> len(allpos_lst)
    2

    Parameters
    ----------
    atomgrid: List[LebedevGrid]
        List of Lebedev grids for each atom. (natoms)
    atompos: torch.Tensor
        Position of each atom. (natoms, ndim)

    Returns
    -------
    Tuple[List[torch.Tensor], torch.Tensor, torch.Tensor]
        List of grid positions for each atom, the concatenated grid
        positions, and the dV elements.

    """
    allpos_lst = [
        # rgrid: (ngrid[i], ndim), pos: (ndim,)
        (gr.get_rgrid() + pos) \
        for (gr, pos) in zip(atomgrid, atompos)]
    rgrid = torch.cat(allpos_lst, dim=0)  # (ngrid, ndim)

    # calculate the dvol for an isolated atom
    dvol_atoms = torch.cat([gr.get_dvolume() for gr in atomgrid], dim=0)

    return allpos_lst, rgrid, dvol_atoms


def _get_atom_weights(rgrids: List[torch.Tensor],
                      atompos: torch.Tensor,
                      atomradii: Optional[torch.Tensor] = None,
                      ratom_adjust: str = "becke") -> torch.Tensor:
    """Calculate the weights for each grid point due to the atoms

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LebedevGrid, RadialGrid
    >>> from deepchem.utils.dft_utils.grid.multiatoms_grid import _get_atom_weights
    >>> grid = RadialGrid(100, grid_integrator="chebyshev", grid_transform="logm3")
    >>> atomgrid = [LebedevGrid(grid, 3), LebedevGrid(grid, 3)]
    >>> atompos = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
    >>> rgrids, _, _ = _construct_rgrids(atomgrid, atompos)
    >>> w = _get_atom_weights(rgrids, atompos)
    >>> w.shape
    torch.Size([1200])

    Parameters
    ----------
    rgrids: List[torch.Tensor]
        List of grid positions for each atom, with length natoms
        consisting of absolute position of the grids (natoms)
    atompos: torch.Tensor
        Position of each atom. (natoms, ndim)
    atomradii: Optional[torch.Tensor] (default None)
        Radii of each atom. (natoms,) or None
    ratom_adjust: str (default "becke")
        Adjustment method for the atom radii. Available: ['becke', 'treutler']

    Returns
    -------
    torch.Tensor
        The weights for each grid point. (ngrid,)

    """
    assert len(rgrids) == atompos.shape[0]
    dtype = atompos.dtype
    device = atompos.device

    natoms = atompos.shape[0]
    rdatoms = atompos - atompos.unsqueeze(1)  # (natoms, natoms, ndim)
    # add the diagonal to stabilize the gradient calculation
    rdatoms = rdatoms + torch.eye(rdatoms.shape[0],
                                  dtype=rdatoms.dtype,
                                  device=rdatoms.device).unsqueeze(-1)
    ratoms = torch.norm(rdatoms, dim=-1)  # (natoms, natoms)

    # calculate the distortion due to heterogeneity
    # (Appendix in Becke's https://doi.org/10.1063/1.454033)
    if atomradii is not None:
        if ratom_adjust == "becke":
            rad = atomradii
        elif ratom_adjust == "treutler":
            # https://aip.scitation.org/doi/pdf/10.1063/1.469408 eq (13)
            rad = atomradii**0.5
        else:
            msg = "Unknown atom adjustment: %s. Available: ['becke', 'treutler']" % ratom_adjust
            raise ValueError(msg)
        # chiij = rad / rad.unsqueeze(1)  # (natoms, natoms)
        uij = (rad - rad.unsqueeze(1)) / \
              (rad + rad.unsqueeze(1))
        aij = torch.clamp(uij / (uij * uij - 1), min=-0.45,
                          max=0.45)  # (natoms, natoms)
        aij = aij.unsqueeze(-1)  # (natoms, natoms, 1)

    xyz_full = torch.cat(rgrids, dim=0)
    # xyz_full: (ngrid, ndim)
    # cdist is more efficient but produces nan in second grad
    # rgatoms = torch.cdist(atompos, xyz, p=2.0)  # (natoms, ngrid)
    w_list: List[torch.Tensor] = []
    ioff = 0
    for ia in range(natoms):
        # concatenate the grid to save memory
        iend = ioff + rgrids[ia].shape[0]
        xyz = xyz_full[ioff:iend, :]

        rgatoms = torch.norm(xyz - atompos.unsqueeze(1),
                             dim=-1)  # (natoms, ngrid)
        mu_ij = (rgatoms - rgatoms.unsqueeze(1))  # (natoms, natoms, ngrid)
        mu_ij /= ratoms.unsqueeze(-1)  # (natoms, natoms, ngrid)

        if atomradii is not None:
            # mu_ij += aij * (1 - mu_ij * mu_ij)
            mu_ij2 = mu_ij * mu_ij
            mu_ij2 -= 1
            mu_ij2 *= (-aij)
            mu_ij2 += mu_ij
            mu_ij = mu_ij2

        # making mu_ij sparse for efficiency
        # threshold: mu_ij < 0.65 (s > 1e-3), mu_ij < 0.74 (s > 1e-4)
        nnz_idx_bool = torch.all(mu_ij < 0.74,
                                 dim=0).unsqueeze(0)  # (1, natoms, ngrid)
        mu_ij_nnz = mu_ij[:, nnz_idx_bool.squeeze(0)].reshape(
            -1)  # (natoms * nnz_col)
        nnz_idx0 = torch.nonzero(nnz_idx_bool).unsqueeze(0)  # (1, nnz_col, 3)
        nnz_col_idx = nnz_idx0[0, :, 1:].transpose(-2, -1)  # (2, nnz_col)
        nnz_idx_atm = torch.zeros((natoms, 1, 3),
                                  dtype=nnz_idx0.dtype,
                                  device=nnz_idx0.device)  # (natoms, 1, 3)
        nnz_idx_atm[:, 0, 0] = torch.arange(natoms,
                                            dtype=nnz_idx0.dtype,
                                            device=nnz_idx0.device)
        nnz_idx = (nnz_idx0 + nnz_idx_atm).reshape(
            -1, 3)  # (nnz = natoms * nnz_col, 3)
        atom_diag_idx = nnz_idx[:, 0] == nnz_idx[:, 1]  # (nnz,)

        f = mu_ij_nnz
        for _ in range(3):
            # f = 0.5 * f * (3 - f * f)
            f2 = f.clone()
            f2 *= f
            f2 -= 3
            f2 *= f
            f2 *= (-0.5)
            f = f2

        # small epsilon to avoid nan in the gradient
        # s = 0.5 * (1. + 1e-12 - f)  # (nnz,)
        s = f
        s -= (1 + 1e-12)
        s *= (-0.5)

        # s += 0.5 * torch.eye(natoms)
        s[atom_diag_idx] += 0.5
        s = s.reshape(natoms, -1)  # (natoms, nnz_col)
        psparse = s.prod(dim=0)  # (nnz_col,)

        # densify and normalize p
        p = torch.zeros((natoms, mu_ij.shape[-1]), dtype=dtype,
                        device=device)  # (natoms, ngrid)
        p[nnz_col_idx[0], nnz_col_idx[1]] = psparse
        p = p / p.sum(dim=0, keepdim=True)  # (natoms, ngrid)

        # save the grid
        w_list.append(p[ia])
        ioff = iend

    w = torch.cat(w_list, dim=-1)  # (ngrid)
    return w



# flake8: noqa
import logging

logger = logging.getLogger(__name__)
try:
    from deepchem.metalearning.maml import MAML
    from deepchem.metalearning import MAML
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading some PyTorch models, missing a dependency. {e}')


class MetaLearner(object):
    """Model and data to which the MAML algorithm can be applied.

    To use MAML, create a subclass of this defining the learning problem to solve.
    It consists of a model that can be trained to perform many different tasks, and
    data for training it on a large (possibly infinite) set of different tasks.
    """

    def compute_model(self, inputs, variables, training):
        """Compute the model for a set of inputs and variables.

        Parameters
        ----------
        inputs: list of tensors
            the inputs to the model
        variables: list of tensors
            the values to use for the model's variables.  This might be the actual
            variables (as returned by the MetaLearner's variables property), or
            alternatively it might be the values of those variables after one or more
            steps of gradient descent for the current task.
        training: bool
            indicates whether the model is being invoked for training or prediction

        Returns
        -------
        (loss, outputs) where loss is the value of the model's loss function, and
        outputs is a list of the model's outputs
        """
        raise NotImplementedError("Subclasses must implement this")

    @property
    def variables(self):
        """Get the list of variables to train."""
        raise NotImplementedError("Subclasses must implement this")

    def select_task(self):
        """Select a new task to train on.

        If there is a fixed set of training tasks, this will typically cycle through them.
        If there are infinitely many training tasks, this can simply select a new one each
        time it is called.
        """
        raise NotImplementedError("Subclasses must implement this")

    def get_batch(self):
        """Get a batch of data for training.

        This should return the data as a list of arrays, one for each of the model's
        inputs.  This will usually be called twice for each task, and should
        return a different batch on each call.
        """
        raise NotImplementedError("Subclasses must implement this")

    def parameters(self):
        """Returns an iterator over the MetaLearner parameters."""
        raise NotImplementedError("Subclasses must implement this")



"""
A fake data generator
"""
import random
import numpy as np
from deepchem.data import NumpyDataset
from deepchem.feat import GraphData


class FakeGraphGenerator:
    """Generates a random graphs which can be used for testing or other purposes.

    The generated graph supports both node-level and graph-level labels.

    Example
    -------
    >>> from deepchem.utils.fake_data_generator import FakeGraphGenerator
    >>> fgg  = FakeGraphGenerator(min_nodes=8, max_nodes=10,  n_node_features=5, avg_degree=8, n_edge_features=3, n_classes=2, task='graph', z=5)
    >>> graphs = fgg.sample(n_graphs=10)
    >>> type(graphs)
    <class 'deepchem.data.datasets.NumpyDataset'>
    >>> type(graphs.X[0])
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> len(graphs) == 10  # num_graphs
    True

    Note
    ----
    The FakeGraphGenerator class is based on torch_geometric.dataset.FakeDataset
    class.
    """

    def __init__(self,
                 min_nodes: int = 10,
                 max_nodes: int = 10,
                 n_node_features: int = 5,
                 avg_degree: int = 4,
                 n_edge_features: int = 3,
                 n_classes: int = 2,
                 task: str = 'graph',
                 **kwargs):
        """
        Parameters
        ----------
        min_nodes: int, default 10
            Minimum number of permissible nodes in a graph
        max_nodes: int, default 10
            Maximum number of permissible nodes in a graph
        n_node_features: int, default 5
            Average number of node features in a graph
        avg_degree: int, default 4
            Average degree of the graph (avg_degree should be a positive number greater than the min_nodes)
        n_edge_features: int, default 3
            Average number of features in the edge
        task: str, default 'graph'
            Indicates node-level labels or graph-level labels
        kwargs: optional
            Additional graph attributes and their shapes , e.g. `global_features = 5`
        """
        assert avg_degree >= 1, "Average degree should be greater than 0"
        self.min_nodes = min_nodes
        self.max_nodes = max_nodes
        self.avg_degree = avg_degree
        self.n_node_features = n_node_features
        self.n_edge_features = n_edge_features
        self.n_classes = n_classes
        self.task = task
        self.kwargs = kwargs

    def sample(self, n_graphs: int = 100) -> NumpyDataset:
        """Samples graphs

        Parameters
        ----------
        n_graphs: int, default 100
            Number of graphs to generate

        Returns
        -------
        graphs: NumpyDataset
            Generated Graphs
        """
        graphs, labels = [], []
        for i in range(n_graphs):
            n_nodes = random.randint(self.min_nodes, self.max_nodes)
            edge_index = generate_edge_index(n_nodes, self.avg_degree)
            n_edges = edge_index.shape[1]

            if self.task == 'graph':
                graph_label = random.randint(0, self.n_classes - 1)
                node_features = np.random.rand(
                    n_nodes, self.n_node_features) + graph_label
                edge_features = np.random.rand(
                    n_edges, self.n_edge_features) + graph_label
                kwargs = {}
                for feature_name, feature_shape in self.kwargs.items():
                    kwargs[feature_name] = np.random.rand(
                        1, feature_shape) + graph_label
                labels.append(graph_label)

            elif self.task == 'node':
                node_label = np.random.randint(0, self.n_classes - 1,
                                               n_nodes).astype(np.float64)
                node_features = np.random.rand(
                    n_nodes, self.n_node_features) + node_label.reshape(-1, 1)
                # For a node-prediction task, label is not added to edge features and other global features
                # because label here is a node-level attribute and not a graph-level attribute
                edge_features = np.random.rand(n_edges, self.n_edge_features)
                kwargs = {}
                for feature_name, feature_shape in self.kwargs.items():
                    kwargs[feature_name] = np.random.rand(1, feature_shape)
                kwargs['y'] = node_label

            graph = GraphData(node_features, edge_index, edge_features,
                              **kwargs)
            graphs.append(graph)

            if self.task == 'graph':
                graph_dataset = NumpyDataset(X=np.array(graphs),
                                             y=np.array(labels))
            elif self.task == 'node':
                # In this case, the 'y' attribute of GraphData will contain the
                # node-level labels.
                graph_dataset = NumpyDataset(X=np.array(graphs))
        return graph_dataset


def generate_edge_index(n_nodes: int,
                        avg_degree: int,
                        remove_loops: bool = True) -> np.ndarray:
    """Returns source and destination nodes for `num_nodes * avg_degree` number of randomly
    generated edges. If remove_loops is True, then self-loops from the edge_index pairs
    are removed.

    Parameters
    ----------
    n_nodes: int
        Number of nodes in the graph
    avg_degree: int
        Average degree per node in a graph
    remove_loops: bool
        Remove self-loops in a graph
    """
    n_edges = n_nodes * avg_degree
    edge_index = np.random.randint(low=0, high=n_nodes, size=(2, n_edges))

    if remove_loops:
        edge_index = remove_self_loops(edge_index)
    return edge_index


def remove_self_loops(edge_index: np.ndarray) -> np.ndarray:
    """Removes self-loops from a given set of edges

    Parameters
    ----------
    edge_index: numpy.ndarray
        An numpy array of shape (2, |num_edges|) representing edges in a graph
    """
    mask = []
    for i in range(edge_index.shape[1]):
        if edge_index[0][i] != edge_index[1][i]:
            # not a self-loop
            mask.append(i)
    return edge_index[:, mask]



# flake8:noqa
import logging

logger = logging.getLogger(__name__)

from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.torch_models.modular import ModularTorchModel
from deepchem.models.torch_models.attentivefp import AttentiveFP, AttentiveFPModel
from deepchem.models.torch_models.cgcnn import CGCNN, CGCNNModel
from deepchem.models.torch_models.gat import GAT, GATModel
from deepchem.models.torch_models.gcn import GCN, GCNModel
from deepchem.models.torch_models.gan import GAN, GANModel, WGANModel, GradientPenaltyLayer
from deepchem.models.torch_models.molgan import BasicMolGANModel
from deepchem.models.torch_models.infograph import InfoGraphStar, InfoGraphStarModel, InfoGraphEncoder, GINEncoder, InfoGraph, InfoGraphModel, InfoGraphEncoder
from deepchem.models.torch_models.mpnn import MPNN, MPNNModel
from deepchem.models.torch_models.lcnn import LCNN, LCNNModel
from deepchem.models.torch_models.pagtn import Pagtn, PagtnModel
from deepchem.models.torch_models.mat import MAT, MATModel
from deepchem.models.torch_models.megnet import MEGNetModel
from deepchem.models.torch_models.normalizing_flows_pytorch import NormalizingFlow
from deepchem.models.torch_models.layers import MultilayerPerceptron, CNNModule, CombineMeanStd, WeightedLinearCombo, AtomicConvolution, NeighborList, SetGather, EdgeNetwork, WeaveLayer, WeaveGather, MolGANConvolutionLayer, MolGANAggregationLayer, MolGANMultiConvolutionLayer, MolGANEncoderLayer, VariationalRandomizer, EncoderRNN, DecoderRNN, AtomicConv, GraphConv, GraphPool, GraphGather, DAGGather, DAGLayer, Fiber
from deepchem.models.torch_models.cnn import CNN
from deepchem.models.torch_models.scscore import ScScore, ScScoreModel
from deepchem.models.torch_models.weavemodel_pytorch import Weave, WeaveModel
from deepchem.models.torch_models.attention import ScaledDotProductAttention, SelfAttention
from deepchem.models.torch_models.grover import GroverModel, GroverPretrain, GroverFinetune
from deepchem.models.torch_models.readout import GroverReadout
from deepchem.models.torch_models.dtnn import DTNN, DTNNModel
from deepchem.models.torch_models.seqtoseq import SeqToSeq, SeqToSeqModel
from deepchem.models.torch_models.acnn import AtomConvModel
from deepchem.models.torch_models.progressive_multitask import ProgressiveMultitask, ProgressiveMultitaskModel, ProgressiveMultitaskClassifier, ProgressiveMultitaskRegressor
from deepchem.models.torch_models.text_cnn import TextCNNModel
from deepchem.models.torch_models.flows import Flow, Affine, MaskedAffineFlow, ActNorm, ClampExp, ConstScaleLayer, MLPFlow, NormalizingFlow, NormalizingFlowModel
from deepchem.models.torch_models.unet import UNet, UNetModel
from deepchem.models.torch_models.graphconvmodel import _GraphConvTorchModel, GraphConvModel
from deepchem.models.torch_models.smiles2vec import Smiles2Vec, Smiles2VecModel
from deepchem.models.torch_models.robust_multitask import RobustMultitask
from deepchem.models.torch_models.lstm_generator_models import LSTMNeuralNet, LSTMGenerator
from deepchem.models.torch_models.hf_models import HuggingFaceModel
from deepchem.models.torch_models.inceptionv3 import InceptionV3Model, InceptionA, InceptionB, InceptionC, InceptionD, InceptionE, InceptionAux, BasicConv2d
from deepchem.models.torch_models.mobilenetv2 import MobileNetV2Model, InvertedResidual
from deepchem.models.torch_models.robust_multitask import RobustMultitask, RobustMultitaskClassifier, RobustMultitaskRegressor
from deepchem.models.torch_models.IRV import IRVLayer, MultitaskIRVClassifier
from deepchem.models.torch_models.pinns_model import PINNModel
from deepchem.models.torch_models.chemnet_layers import Stem
try:
    from deepchem.models.torch_models.dmpnn import DMPNN, DMPNNModel
    from deepchem.models.torch_models.gnn import GNN, GNNHead, GNNModular
    from deepchem.models.torch_models.pna_gnn import AtomEncoder, BondEncoder, PNALayer, PNAGNN, PNA
    from deepchem.models.torch_models.gnn3d import Net3D, InfoMax3DModular
    from deepchem.models.torch_models.weavemodel_pytorch import Weave, WeaveModel
    from deepchem.models.torch_models.mxmnet import MXMNet
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading modules with pytorch-geometric dependency, missing a dependency. {e}'
    )
try:
    from deepchem.models.torch_models.hf_models import HuggingFaceModel
    from deepchem.models.torch_models.chemberta import Chemberta
    from deepchem.models.torch_models.molformer import MoLFormer
    from deepchem.models.torch_models.prot_bert import ProtBERT
    from deepchem.models.torch_models.oneformer import OneFormer

except ModuleNotFoundError as e:
    logger.warning(f'Skipped loading modules with transformers dependency. {e}')



from typing import Optional, Tuple
from rdkit import Chem
import re
from rdkit.Chem import rdmolops

ALLOWED_CONVERSION_TYPES = ["alternate", "block", "random"]


class PSMILES2WDGConverter:
    """
    This class is used to convert a PSMILES string to a to corresponding WDGraph string.
    The conversion can be done in two ways:
        1. With metadata: The metadata contains the indices of the atoms to be bonded and the residue.
        2. Without metadata: The conversion is done using the conversion types specified in the constructor.
    The mechanism is as follows:
        1. The PSMILES string is split into two parts.
        2. The bond is broken between the specified indices.
        3. The fragments are converted to SMILES strings.
        4. The SMILES strings are combined with the residue to form the final WDGraph string.
    This specific mechanism is utilized to convert a PSMILES string to a WDGraph string such that the representational
    variation is tested with neural network architectural differences in our research paper "Open-Source Polymer Generative Pipeline" [1]_.

    References
    ----------
    .. [1] Mohanty, Debasish, et al. "Open-source Polymer Generative Pipeline."
        arXiv preprint arXiv:2412.08658 (2024).

    Examples
    --------
    >>> from rdkit import Chem
    >>> from deepchem.utils.poly_converters import PSMILES2WDGConverter
    >>> psmiles = "*CCCC*"
    >>> metadata = {
    ...     "indicies": [2, 3],
    ...     "seq_index": 4,
    ...     "residue": "0.5|0.5|",
    ...     "smiles_type": "SMARTS"
    ... }
    >>> converter = PSMILES2WDGConverter()
    >>> result = converter([psmiles], [metadata])
    >>> print(result)
    [['[*:1]CC.[*:2]C[*:3][*:4]C|0.5|0.5|']]
    """

    def __init__(self, conversion_types: list = ["alternate"]) -> None:
        """
        Parameters
        ----------
        conversion_types: list
            The conversion types to be used for the conversion. The default is ["alternate"].
        """
        for conversion in conversion_types:
            if conversion not in ALLOWED_CONVERSION_TYPES:
                raise ValueError(
                    f"The conversion type {conversion} is not supported choose any from {ALLOWED_CONVERSION_TYPES}"
                )
        self.conversion_types = conversion_types
        self.CONVERT_OUTRO_MAPPER = {
            "alternate":
                "<1-3:0.5:0.5<1-4:0.5:0.5<2-3:0.5:0.5<2-4:0.5:0.5",
            "block":
                "<1-2:0.75:0.75<3-4:0.75:0.75<1-3:0.125:0.125<1-4:0.125:0.125<2-3:0.125:0.125<2-4:0.125:0.125",
            "random":
                "<1-2:0.50:0.50<3-4:0.50:0.50<1-3:0.33:0.33<1-4:0.33:0.33<2-3:0.33:0.33<2-4:0.33:0.33"
        }

    def index_wildcards(self, psmiles_part: str) -> str:
        """
        This function is used to index the wildcard atoms in the PSMILES string.

        Parameters
        ----------
        psmiles_part: str
            The PSMILES string to be indexed.

        Returns
        -------
        str
            The indexed PSMILES string.
        """
        counter = 1
        mod_psmiles = ""
        for idx, char in enumerate(psmiles_part):
            if char == "*":
                mod_psmiles += "[" + str(counter) + "*" + "]"
                counter += 1
            else:
                mod_psmiles += char
        return mod_psmiles

    def add_indicies_to_smiles_from_meta(self, smiles: str,
                                         seq_index: int) -> str:
        """
        This function is used to add the indices to the SMILES string from the metadata.

        Parameters
        ----------
        smiles: str
            The SMILES string to be modified.
        seq_index: int
            The index of the sequence in the SMILES string.

        Returns
        -------
        str
            The modified SMILES string.
        """
        restored_smiles = smiles[:seq_index] + "**" + smiles[seq_index:]
        return restored_smiles

    def make_wdgraph_string_from_meta(self, psmiles: str,
                                      indicies: list) -> str:
        """
        This function is used to make the WDGraph string from the metadata.

        Parameters
        ----------
        psmiles: str
            The PSMILES string to be converted.
        indicies: list
            The indices of the atoms to be bonded.

        Returns
        -------
        str
            The WDGraph string.
        """
        # Convert the combined SMILES to an RWMol object for editing
        combined_mol = Chem.MolFromSmiles(psmiles)
        editable_mol = Chem.RWMol(combined_mol)

        # Break the bond between the specified indices
        editable_mol.RemoveBond(indicies[0], indicies[1])

        # Generate fragments (separate molecules) after breaking the bond
        fragments = Chem.GetMolFrags(editable_mol, asMols=True)

        # Convert each fragment to its respective SMILES string
        fragment_smiles = [Chem.MolToSmiles(frag) for frag in fragments]
        return ".".join(fragment_smiles)

    def convert_smiles_to_SMARTS(self, smiles_string: str) -> str:
        """
        This function is used to convert the SMILES string to SMARTS string.

        Parameters
        ----------
        smiles_string: str
            The SMILES string to be converted.

        Returns
        -------
        str
            The converted SMARTS string.
        """
        counter = 1
        mod_string = ""
        for s in smiles_string:
            if s == "*":
                mod_string += f"[*:{counter}]"
                counter += 1
            else:
                mod_string += s
        return mod_string

    def convert_smiles_to_MOD_SMARTS(self, smiles_string: str) -> str:
        """
        This function is used to convert the SMILES string to MOD SMARTS string.

        Parameters
        ----------
        smiles_string: str
            The SMILES string to be converted.

        Returns
        -------
        str
            The converted MOD SMARTS string.
        """
        counter = 1
        mod_string = ""
        for s in smiles_string:
            if s == "*":
                mod_string += f"[{counter}*]"
                counter += 1
            else:
                mod_string += s
        return mod_string

    def compose_from_meta(self, psmiles: str, metadata: dict) -> str:
        """
        This function is used to compose the WDGraph string from the metadata.

        Parameters
        ----------
        psmiles: str
            The PSMILES string to be converted.
        metadata: dict
            The metadata containing the indices of the atoms to be bonded, the residue, and the type of SMILES string.

        Returns
        -------
        str
            The composed WDGraph string.
        """
        w_idx_psmiles = self.add_indicies_to_smiles_from_meta(
            psmiles, metadata["seq_index"])
        smile_part = self.make_wdgraph_string_from_meta(w_idx_psmiles,
                                                        metadata["indicies"])
        if metadata["smiles_type"] == "SMARTS":
            mod_smiles = self.convert_smiles_to_SMARTS(smile_part)
        elif metadata["smiles_type"] == "MOD_SMARTS":
            mod_smiles = self.convert_smiles_to_MOD_SMARTS(smile_part)
        return mod_smiles + "|" + metadata["residue"]

    def convert(self,
                psmiles_string: str,
                metadata: Optional[dict] = None) -> list:
        """
        This function is used to convert the PSMILES string to a WDGraph string.

        Parameters
        ----------
        psmiles_string: str
            The PSMILES string to be converted.
        metadata: dict
            The metadata containing the indices of the atoms to be bonded, the residue, and the type of SMILES string.

        Returns
        -------
        list
            The converted WDGraph string.
        """
        converted = []
        if metadata:
            converted.append(self.compose_from_meta(psmiles_string, metadata))
        else:
            intro = self.index_wildcards(psmiles_string + "." + psmiles_string)
            midtro = "|0.5|0.5|"
            for conversion_type in self.conversion_types:
                outro = self.CONVERT_OUTRO_MAPPER[conversion_type]
                converted.append(intro + midtro + outro)
        return converted

    def __call__(self,
                 psmiles_list: list,
                 metadata_list: Optional[list] = None) -> list:
        """
        This function is used to call the conversion process.

        Parameters
        ----------
        psmiles_list: list
            The list of PSMILES strings to be converted.
        metadata_list: list
            The list of metadata containing the indices of the atoms to be bonded, the residue, and the type of SMILES string.

        Returns
        -------
        list
            The converted WDGraph strings.
        """
        converted_list = []
        if metadata_list is None:
            for psmiles_string in psmiles_list:
                converted_list.append(self.convert(psmiles_string))
        else:
            for psmiles_string, metadata in zip(psmiles_list, metadata_list):
                converted_list.append(self.convert(psmiles_string, metadata))
        return converted_list


class WDG2PSMILESConverter:
    """
    This class is used to convert a WDGraph string to a corresponding PSMILES string.
    The mechanism returns two formats:
        1. With metadata: Gives metadata containing the indices of the atoms to be bonded and the residue.
        2. Without metadata: Gives only the converted PSMILES string.
           (the exact reverse conversion of PSMILES to WDG will be done considering alternating structure only)
    The mechanism is as follows:
        1. The WDGraph string is split into two parts (PSMILES part, Residue part)
        2. The PSMILES part is converted from SMART notation to normal notation.
        3. The PSMILES part is converted to a single molecule keeping the bond breaking indicies as metadata.
        4. The final PSMILES is returned with the metadata of the bond breaking indices and initial residue.
    This specific mechanism is utilized to convert a PSMILES string to a WDGraph string such that the representational
    variation is tested with neural network architectural differences in our research paper "Open-Source Polymer Generative Pipeline" [1]_.

    References
    ----------
    .. [1] Mohanty, Debasish, et al. "Open-source Polymer Generative Pipeline."
        arXiv preprint arXiv:2412.08658 (2024).

    Examples
    --------
    >>> from rdkit import Chem
    >>> from deepchem.utils.poly_converters import WDG2PSMILESConverter
    >>> wd_graph_string = "[*:1]CC[*:2].[*:3]CC[*:4]|0.5|0.5|<1-3:0.5:0.5<1-4:0.5:0.5<2-3:0.5:0.5<2-4:0.5:0.5"
    >>> converter = WDG2PSMILESConverter(return_metadata = True)
    >>> result, meta = converter([wd_graph_string])
    >>> print(result)
    ['*CCCC*']
    >>> print(meta)
    [{'indicies': [3, 4], 'seq_index': 3, 'residue': '0.5|0.5|<1-3:0.5:0.5<1-4:0.5:0.5<2-3:0.5:0.5<2-4:0.5:0.5', 'smiles_type': 'SMARTS'}]
    """

    def __init__(self, return_metadata: bool = True) -> None:
        """
        Parameters
        ----------
        return_metadata: bool
            Whether to return metadata or not. The default is True.
        """
        self.return_metadata = return_metadata

    def get_wildcard_bond_indecies(self,
                                   combined_psmiles: str) -> Tuple[int, int]:
        """
        This function is used to get the wildcard bond indices in the combined PSMILES string.

        Parameters
        ----------
        combined_psmiles: str
            The combined PSMILES string.

        Returns
        -------
        Tuple(int, int)
            The wildcard bond indices.
        """
        if "**" not in combined_psmiles:
            raise ValueError(
                "The combined PSMILES must have two wildcard atoms together !")
        atoms = [
            atom.GetSymbol()
            for atom in Chem.MolFromSmiles(combined_psmiles).GetAtoms()
        ]
        found = False
        for i, atom in enumerate(atoms):
            if atom == "*" and found:
                return i, i + 1
            elif atom == "*":
                found = True
        return -1, -1

    def convert_smiles_part(self, smiles: str) -> Tuple[str, int, int, int]:
        """
        This function is used to convert the SMILES part of the WDG string to a PSMILES string and returns the indicies along side.

        Parameters
        ----------
        smiles: str
            The SMILES part of the WDG string.

        Returns
        -------
        Tuple(str, int, int, int)
            The converted PSMILES string, the bond breaking indicies, and the sequence index.
        """
        # Split the SMILES string into individual molecules
        if '.' not in smiles:
            raise ValueError(
                "The input SMILES string does not contain a joining notation")

        mols = [Chem.MolFromSmiles(s) for s in smiles.split('.')]

        # Make sure both molecules are valid
        if None in mols:
            raise ValueError("One of the SMILES strings is invalid")

        # Combine the two molecules into a single editable molecule
        combined_mol = Chem.CombineMols(mols[0], mols[1])
        combined_mol = Chem.RWMol(combined_mol)

        # Get atom indices of wildcard '*' atoms in the combined molecule
        wildcard_indices = [
            atom.GetIdx()
            for atom in combined_mol.GetAtoms()
            if atom.GetSymbol() == '*'
        ]

        # Check if there are at least two wildcards to form a bond
        if len(wildcard_indices) < 2:
            raise ValueError("Not enough wildcards to form a bond")

        # Select the indices of the wildcards you want to bond
        # For example, let's bond the second and third wildcards (index 1 and 2)
        # Note: Adjust indices as per your needs
        index1 = wildcard_indices[1]
        index2 = wildcard_indices[2]

        # Add a single bond between the selected wildcard atoms
        combined_mol.AddBond(index1, index2, Chem.BondType.SINGLE)

        # Remove the wildcard atoms after bonding
        rdmolops.DeleteSubstructs(combined_mol, Chem.MolFromSmiles('*'))

        # Convert the RWMol back to a Mol object
        final_mol = combined_mol.GetMol()

        # Generate the SMILES string of the final combined molecule
        final_smiles = Chem.MolToSmiles(final_mol)
        index_1, index_2 = self.get_wildcard_bond_indecies(final_smiles)
        seq_index = final_smiles.find("**")
        return final_smiles.replace("**", ""), index_1, index_2, seq_index

    def replace_SMARTS(self, smiles: str) -> str:
        """
        This function is used to replace the SMARTS notation in the SMILES string.

        Parameters
        ----------
        smiles: str
            The SMILES string to be modified.

        Returns
        -------
        str
            The modified SMILES string
        """
        pattern = r"\[\*:\d+\]"
        result = re.sub(pattern, "*", smiles)
        return result

    def replace_mod_SMARTS(self, mod_smiles: str) -> str:
        """
        This function is used to replace the MOD SMARTS notation in the SMILES string.

        Parameters
        ----------
        mod_smiles: str
            The SMILES string to be modified.

        Returns
        -------
        str
            The modified SMILES string.
        """
        pattern = r"\[\d+\*\]"
        result = re.sub(pattern, "*", mod_smiles)
        return result

    def convert(self, wdg_graph_string: str) -> Tuple[str, dict]:
        """
        This is used to convert a single WDGraph string to a PSMILES string and form the metadata.

        Parameters
        ----------
        wdg_graph_string: str
            The WDGraph string to be converted.

        Returns
        -------
        Tuple(str, dict)
            The converted PSMILES string and the metadata.
        """
        parts = wdg_graph_string.split("|")
        monomer_smiles = parts[0]
        smiles_type = None
        residue = "|".join(parts[1:])
        if "*:" in monomer_smiles:
            mod_smiles = self.replace_SMARTS(monomer_smiles)
            smiles_type = "SMARTS"
        elif "*]" in monomer_smiles:
            mod_smiles = self.replace_mod_SMARTS(monomer_smiles)
            smiles_type = "MOD_SMARTS"
        else:
            raise ValueError(
                f"The data {wdg_graph_string} is not valid for this conversion !"
            )
        psmiles, index_1, index_2, seq_index = self.convert_smiles_part(
            mod_smiles)
        if self.return_metadata:
            return psmiles, {
                "indicies": [index_1, index_2],
                "seq_index": seq_index,
                "residue": residue,
                "smiles_type": smiles_type
            }
        else:
            return psmiles, {}

    def __call__(self, wd_graph_list: list) -> Tuple[list, list]:
        """
        This function is used to call the conversion process.

        Parameters
        ----------
        wd_graph_list: list
            The list of WDGraph strings to be converted.

        Returns
        -------
        Tuple(list, list)
            The converted PSMILES strings and the metadata.
        """
        converted_psmiles_list = []
        metadata_list = []
        for wd_graph_string in wd_graph_list:
            if self.return_metadata:
                psmiles, metadata = self.convert(wd_graph_string)
                converted_psmiles_list.append(psmiles)
                metadata_list.append(metadata)
            else:
                psmiles, _ = self.convert(wd_graph_string)
                converted_psmiles_list.append(psmiles)
        if self.return_metadata:
            return converted_psmiles_list, metadata_list
        return converted_psmiles_list, []



"""
Topological fingerprints.
"""
from typing import Dict

import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer


class CircularFingerprint(MolecularFeaturizer):
    """Circular (Morgan) fingerprints.

    Extended Connectivity Circular Fingerprints compute a bag-of-words style
    representation of a molecule by breaking it into local neighborhoods and
    hashing into a bit vector of the specified size. It is used specifically
    for structure-activity modelling. See [1]_ for more details.

    References
    ----------
    .. [1] Rogers, David, and Mathew Hahn. "Extended-connectivity fingerprints."
        Journal of chemical information and modeling 50.5 (2010): 742-754.

    Note
    ----
    This class requires RDKit to be installed.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> smiles = ['C1=CC=CC=C1']
    >>> # Example 1: (size = 2048, radius = 4)
    >>> featurizer = dc.feat.CircularFingerprint(size=2048, radius=4)
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (2048,)

    >>> # Example 2: (size = 2048, radius = 4, sparse = True, smiles = True)
    >>> featurizer = dc.feat.CircularFingerprint(size=2048, radius=8,
    ...                                          sparse=True, smiles=True)
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0]) # dict containing fingerprints
    <class 'dict'>

    """

    def __init__(self,
                 radius: int = 2,
                 size: int = 2048,
                 chiral: bool = False,
                 bonds: bool = True,
                 features: bool = False,
                 sparse: bool = False,
                 smiles: bool = False,
                 is_counts_based: bool = False):
        """
        Parameters
        ----------
        radius: int, optional (default 2)
            Fingerprint radius.
        size: int, optional (default 2048)
            Length of generated bit vector.
        chiral: bool, optional (default False)
            Whether to consider chirality in fingerprint generation.
        bonds: bool, optional (default True)
            Whether to consider bond order in fingerprint generation.
        features: bool, optional (default False)
            Whether to use feature information instead of atom information; see
            RDKit docs for more info.
        sparse: bool, optional (default False)
            Whether to return a dict for each molecule containing the sparse
            fingerprint.
        smiles: bool, optional (default False)
            Whether to calculate SMILES strings for fragment IDs (only applicable
            when calculating sparse fingerprints).
        is_counts_based: bool, optional (default False)
            Whether to generates a counts-based fingerprint.

        """
        self.radius = radius
        self.size = size
        self.chiral = chiral
        self.bonds = bonds
        self.features = features
        self.sparse = sparse
        self.smiles = smiles
        self.is_counts_based = is_counts_based

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate circular fingerprint.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            A numpy array of circular fingerprint.

        """
        try:
            from rdkit import Chem, DataStructs
            from rdkit.Chem import rdMolDescriptors
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.sparse:
            info: Dict = {}
            fp = rdMolDescriptors.GetMorganFingerprint(
                datapoint,
                self.radius,
                useChirality=self.chiral,
                useBondTypes=self.bonds,
                useFeatures=self.features,
                bitInfo=info)
            fp = fp.GetNonzeroElements()  # convert to a dict

            # generate SMILES for fragments
            if self.smiles:
                fp_smiles = {}
                for fragment_id, count in fp.items():
                    root, radius = info[fragment_id][0]
                    env = Chem.FindAtomEnvironmentOfRadiusN(
                        datapoint, radius, root)
                    frag = Chem.PathToSubmol(datapoint, env)
                    smiles = Chem.MolToSmiles(frag)
                    fp_smiles[fragment_id] = {'smiles': smiles, 'count': count}
                fp = fp_smiles
        else:
            if self.is_counts_based:
                fp_sparse = rdMolDescriptors.GetHashedMorganFingerprint(
                    datapoint,
                    self.radius,
                    nBits=self.size,
                    useChirality=self.chiral,
                    useBondTypes=self.bonds,
                    useFeatures=self.features)
                fp = np.zeros(
                    (self.size,), dtype=float
                )  # initialise numpy array of zeros (shape: (required size,))
                DataStructs.ConvertToNumpyArray(fp_sparse, fp)
            else:
                fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
                    datapoint,
                    self.radius,
                    nBits=self.size,
                    useChirality=self.chiral,
                    useBondTypes=self.bonds,
                    useFeatures=self.features)
                fp = np.asarray(fp, dtype=float)
        return fp

    def __hash__(self):
        return hash((self.radius, self.size, self.chiral, self.bonds,
                     self.features, self.sparse, self.smiles))

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return self.radius == other.radius and \
               self.size == other.size and \
               self.chiral == other.chiral and \
               self.bonds == other.bonds and \
               self.features == other.features and \
               self.sparse == other.sparse and \
               self.smiles == other.smiles
